"""Functions for preparing datasets for comparison."""

import cftime as cf
import numpy as np
import xarray as xr

from ilamb3 import dataset as dset
from ilamb3.exceptions import TemporalOverlapIssue


def nest_spatial_grids(*args):
    """Return the arguments interpolated to a nested grid.

    In order to avoid loss of information to interpolation, when comparing two or more
    datasets, ILAMB uses nearest neighbor interpolation on a grid composed of the union
    of the lat/lon breaks of all sources. See this
    [figure](https://agupubs.onlinelibrary.wiley.com/doi/full/10.1029/2018MS001354#jame20779-fig-0002)
    and surrounding discussion for more information.

    Parameters
    ----------
    *args
        Any number of `xr.Datasets` to be included in creating a nested grid.

    Returns
    -------
    *args
        The input *args interpolated to the nested grid.

    """

    def _return_breaks(ds: xr.Dataset, dim_name: str):
        dim = ds[dim_name]
        if "bounds" in dim.attrs and dim.attrs["bounds"] in ds:
            dim = ds[dim.attrs["bounds"]]
        else:
            dim = ds.cf.add_bounds(dim_name)[f"{dim_name}_bounds"]
        return dim.to_numpy().flatten()

    # are the arguments all spatial?
    if not all([dset.is_spatial(arg) for arg in args]):
        return args

    # find the union of all the breaks, and then the centroids of this irregular grid
    lat = np.empty(0)
    lon = np.empty(0)
    for arg in args:
        arg = arg.to_dataset() if isinstance(arg, xr.DataArray) else arg
        lat = np.union1d(lat, _return_breaks(arg, dset.get_dim_name(arg, "lat")))
        lon = np.union1d(lon, _return_breaks(arg, dset.get_dim_name(arg, "lon")))
    lat = 0.5 * (lat[:-1] + lat[1:])
    lon = 0.5 * (lon[:-1] + lon[1:])
    out = []
    for arg in args:
        lat_name = dset.get_dim_name(arg, "lat")
        lon_name = dset.get_dim_name(arg, "lon")
        iarg = arg.interp({lat_name: lat, lon_name: lon}, method="nearest")
        # if 'bounds' existed, they will now be interpolated and incorrect
        for dim_name in [lat_name, lon_name]:
            dim = iarg[dim_name]
            try:
                iarg = iarg.drop_vars(dim.attrs["bounds"])
            except Exception:
                pass
        out.append(iarg)
    return out


def is_spatially_aligned(dsa: xr.Dataset, dsb: xr.Dataset) -> bool:
    """Check that the lats and lons of dsa and dsb close to each other.

    Parameters
    ----------
    dsa, dsb
        The datasets to compare.
    """
    alat_name = dset.get_dim_name(dsa, "lat")
    blat_name = dset.get_dim_name(dsb, "lat")
    alon_name = dset.get_dim_name(dsa, "lon")
    blon_name = dset.get_dim_name(dsb, "lon")
    if dsa[alat_name].size != dsb[blat_name].size:
        return False
    if dsa[alon_name].size != dsb[blon_name].size:
        return False
    if not np.allclose(dsa[alat_name], dsb[blat_name]):
        return False
    if not np.allclose(dsa[alon_name], dsb[blon_name]):
        return False
    return True


def pick_grid_aligned(
    ref0: xr.Dataset,
    com0: xr.Dataset,
    ref: xr.Dataset | None = None,
    com: xr.Dataset | None = None,
) -> tuple[xr.Dataset, xr.Dataset]:
    """Return a reference and comparison dataset that is spatially grid-aligned.

    This routine implements some logic to minimize the computational burden of
    interpolating when applying possibly many analysis routines to a given reference and
    comparison.

    - If the original versions are on the same grid, then return them.
    - If interpolated versions are passed in, then check they are in fact the same and
      return.
    - Otherwise compute a nesting of the original versions and return.

    Parameters
    ----------
    ref0
        The original reference dataset
    com0
        The original comparison dataset
    ref
        The optional interpolated reference
    com
        The optional interpolated comparison

    Returns
    -------
    ref, com
        A reference and comparison obtained with minimal additional computational burden

    """
    if is_spatially_aligned(ref0, com0):
        return ref0, com0
    if (ref is not None) and (com is not None):
        if is_spatially_aligned(ref, com):
            return ref, com
    return nest_spatial_grids(ref0, com0)


def trim_time(*args: xr.Dataset, **kwargs: xr.Dataset) -> tuple[xr.Dataset]:
    """Return the datasets trimmed to maximal temporal overlap."""

    def _to_tuple(da: xr.DataArray) -> tuple[int]:
        if da.size != 1:
            raise ValueError("Single element conversions only")
        return (int(da.dt.year), int(da.dt.month))

    def _stamp(ymd: tuple[int]):
        return f"{ymd[0]:4d}-{ymd[1]:02d}"

    # If all the data is monthly, then skip the time bounds in computing the
    # time extents as they tend to mess up the logic with arbitrary calendars.
    # However, if the data is a mean over a long time span (as in biomass or
    # soil carbon), we need the time bounds in the extent to compute the mean of
    # the appropriate quantity.
    time_frequency = [
        float(dset.compute_time_measures(arg).mean().values)
        for arg in list(args) + [arg for _, arg in kwargs.items()]
    ]
    inc_bounds = True
    if np.allclose(time_frequency, 30, atol=3):
        inc_bounds = False
    if np.allclose(time_frequency, 365, atol=3):
        inc_bounds = False

    # Get the time extents in the original calendars
    t0 = []
    tf = []
    for arg in args:
        tbegin, tend = dset.get_time_extent(arg, include_bounds=inc_bounds)
        t0.append(tbegin)
        tf.append(tend)
    for _, arg in kwargs.items():
        tbegin, tend = dset.get_time_extent(arg, include_bounds=inc_bounds)
        t0.append(tbegin)
        tf.append(tend)

    # Convert to a date tuple (year, month, day) and find the maximal overlap
    tmin = max(*[_to_tuple(t) for t in t0])
    tmax = min(*[_to_tuple(t) for t in tf])
    if tmax < tmin:
        raise TemporalOverlapIssue(
            f"Minimum final time {tmax} is after maximum initial time {tmin}"
        )

    # Recast back into native calendar objects and select
    tslice = slice(_stamp(tmin), _stamp(tmax))
    args = list(args)
    for i, arg in enumerate(args):
        args[i] = arg.sel({dset.get_dim_name(arg, "time"): tslice})
    for key, arg in kwargs.items():
        kwargs[key] = arg.sel({dset.get_dim_name(arg, "time"): tslice})

    # Conditional returns based on what was passed in
    if args and not kwargs:
        return args
    if kwargs and not args:
        return kwargs
    return args, kwargs


def same_spatial_grid(
    grid: xr.DataArray | xr.Dataset, *args: xr.Dataset, **kwargs: xr.Dataset
) -> tuple[xr.Dataset]:
    """."""

    def _drop_bounds(ds: xr.Dataset, dim: str) -> xr.Dataset:
        var = ds[dim]
        if "bounds" not in var.attrs:
            return ds
        if var.attrs["bounds"] not in ds:
            return ds
        ds = ds.drop_vars(var.attrs["bounds"])
        ds[dim].attrs.pop("bounds")
        return ds

    args = list(args)
    lat = grid[dset.get_dim_name(grid, "lat")].values
    lon = grid[dset.get_dim_name(grid, "lon")].values
    for i, arg in enumerate(args):
        _, arg = adjust_lon(grid, arg)
        lat_name = dset.get_dim_name(arg, "lat")
        lon_name = dset.get_dim_name(arg, "lon")
        args[i] = arg.interp(
            {lat_name: lat, lon_name: lon},
            method="nearest",
        )
        args[i] = _drop_bounds(args[i], lat_name)
        args[i] = _drop_bounds(args[i], lon_name)
    for key, arg in kwargs.items():
        _, arg = adjust_lon(grid, arg)
        lat_name = dset.get_dim_name(arg, "lat")
        lon_name = dset.get_dim_name(arg, "lon")
        kwargs[key] = arg.interp(
            {lat_name: lat, lon_name: lon},
            method="nearest",
        )
        kwargs[key] = _drop_bounds(kwargs[key], lat_name)
        kwargs[key] = _drop_bounds(kwargs[key], lon_name)

    # Conditional returns based on what was passed in
    if args and not kwargs:
        return args
    if kwargs and not args:
        return kwargs
    return args, kwargs


def adjust_lon(dsa: xr.Dataset, dsb: xr.Dataset) -> tuple[xr.Dataset, xr.Dataset]:
    """When comparing dsb to dsa, we need their longitudes uniformly in
    [-180,180) or [0,360)."""
    alon_name = dset.get_coord_name(dsa, "lon")
    blon_name = dset.get_coord_name(dsb, "lon")
    if alon_name is None or blon_name is None:
        return dsa, dsb
    a360 = (dsa[alon_name].min() >= 0) * (dsa[alon_name].max() <= 360)
    b360 = (dsb[blon_name].min() >= 0) * (dsb[blon_name].max() <= 360)
    if a360 and not b360:
        dsb[blon_name] = dsb[blon_name] % 360
        if "bounds" in dsb[blon_name].attrs and dsb[blon_name].attrs["bounds"] in dsb:
            dsb[dsb[blon_name].attrs["bounds"]] = (
                dsb[dsb[blon_name].attrs["bounds"]] % 360
            )
        dsb = dsb.sortby(blon_name)
    elif not a360 and b360:
        dsb[blon_name] = (dsb[blon_name] + 180) % 360 - 180
        if "bounds" in dsb[blon_name].attrs and dsb[blon_name].attrs["bounds"] in dsb:
            dsb[dsb[blon_name].attrs["bounds"]] = (
                dsb[dsb[blon_name].attrs["bounds"]] + 180
            ) % 360 - 180
        dsb = dsb.sortby(blon_name)
    return dsa, dsb


def handle_timescale_mismatch(
    ref: xr.Dataset, com: xr.Dataset
) -> tuple[xr.Dataset, xr.Dataset]:
    # Initially, we are only going to handle the case where our reference data
    # is a single time entry representing a span significantly larger than a
    # month.
    dt_ref = dset.get_mean_time_frequency(ref)
    dt_com = dset.get_mean_time_frequency(com)
    if np.allclose(dt_ref, dt_com, atol=3):
        return ref, com
    if len(ref[dset.get_dim_name(ref, "time")]) == 1:
        t0, tf = dset.get_time_extent(ref, include_bounds=True)
        com = com.sel(
            {
                dset.get_dim_name(com, "time"): slice(
                    f"{t0.dt.year:04d}-{t0.dt.month:02d}",
                    f"{tf.dt.year:04d}-{tf.dt.month:02d}",
                )
            }
        )
        return ref, com
    raise NotImplementedError(
        f"We encountered a time scale mismatch that we have no logic to handle. {ref} {com}"
    )


def make_comparable(
    ref: xr.Dataset, com: xr.Dataset, varname: str
) -> tuple[xr.Dataset, xr.Dataset]:
    """Return the datasets in a form where they are comparable."""

    # sometimes our data represents a single time step over a long span
    if dset.is_temporal(ref):
        ref, com = handle_timescale_mismatch(ref, com)

    # trim away time, assuming data is monthly
    if dset.is_temporal(ref):
        ref, com = trim_time(ref, com)

    # latlon needs to be 1D arrays
    if dset.is_latlon2d(com[varname]):
        com = dset.latlon2d_to_1d(ref, com[varname]).to_dataset(name=varname)

    # ensure longitudes are uniform
    ref, com = adjust_lon(ref, com)

    # ensure the lat/lon dims are sorted
    if dset.is_spatial(ref):
        ref = ref.sortby([dset.get_dim_name(ref, "lat"), dset.get_dim_name(ref, "lon")])
    if dset.is_spatial(com):
        com = com.sortby([dset.get_dim_name(com, "lat"), dset.get_dim_name(com, "lon")])

    # pick just the sites
    if dset.is_site(ref[varname]):
        com = extract_sites(ref, com, varname)

    # convert units
    com = dset.convert(com, ref[varname].attrs["units"], varname=varname)

    # load into memory
    ref.load()
    com.load()
    return ref, com


def extract_sites(
    ds_site: xr.Dataset, ds_spatial: xr.Dataset, varname: str
) -> xr.Dataset:
    """
    Extract the `ds_site` lat/lons from `ds_spatial choosing the nearest site.

    Parameters
    ----------
    ds_site : xr.Dataset
        The dataset which contains sites.
    ds_spatial : xr.Dataset
        The dataset from which we will make the extraction.
    varname : str
        The name of the variable of interest.

    Returns
    -------
    xr.Dataset
        `ds_spatial` at the sites defined in `da_site`.
    """
    # If this throws an exception, this isn't a site data array
    dset.get_dim_name(ds_site[varname], "site")

    # Get the lat/lon dim names
    lat_site = dset.get_coord_name(ds_site, "lat")
    lon_site = dset.get_coord_name(ds_site, "lon")
    lat_spatial = dset.get_dim_name(ds_spatial, "lat")
    lon_spatial = dset.get_dim_name(ds_spatial, "lon")

    # Store the mean model resolution
    model_res = np.sqrt(
        ds_spatial[lon_spatial].diff(dim=lon_spatial).mean() ** 2
        + ds_spatial[lat_spatial].diff(dim=lat_spatial).mean() ** 2
    )

    # Choose the spatial grid to the nearest site
    ds_spatial = ds_spatial.sel(
        {lat_spatial: ds_site[lat_site], lon_spatial: ds_site[lon_site]},
        method="nearest",
    )

    # Check that these sites are 'close enough'
    dist = np.sqrt(
        (ds_site[lat_site] - ds_spatial[lat_spatial]) ** 2
        + (ds_site[lon_site] - ds_spatial[lon_spatial]) ** 2
    )
    assert (dist < model_res).all()

    # Set these are coordinates though so that we can tell this is site data. We
    # also set them equal to the sites lat/lon so when comparing to other data
    # that the coords stay on
    ds_spatial = ds_spatial.set_coords([lat_spatial, lon_spatial])
    ds_spatial[lat_spatial] = ds_site[lat_site]
    ds_spatial[lon_spatial] = ds_site[lon_site]

    return ds_spatial


def rename_dims(*args):
    """
    Rename the dimension to a uniform canonical name.

    Parameters
    ----------
    *args
        Any number of `xr.Dataset` or `xr.DataArray` objects for which we will change
        the dimension names.

    Returns
    -------
    *args
        The input *args with the dimension names changed.
    """

    def _populate_renames(ds):
        out = {}
        for dim in ["time", "lat", "lon"]:
            try:
                out[dset.get_dim_name(ds, dim)] = dim
            except KeyError:
                pass
        return out

    for arg in args:
        assert isinstance(arg, xr.DataArray | xr.Dataset)
    return [arg.rename(_populate_renames(arg)) for arg in args]


def convert_calendar_monthly_noleap(
    ds: xr.Dataset | xr.DataArray,
) -> xr.Dataset | xr.DataArray:
    """
    Convert the dataset calendar to a monthly noleap.

    Note
    ----
    At some point, we need the index of time to be the same to make comparisons.
    When this is required, we will convert everything assuming that we are
    dealing with monthly data.
    """
    time_name = dset.get_dim_name(ds, "time")
    ds[time_name] = [
        cf.DatetimeNoLeap(t.dt.year, t.dt.month, 15) for t in ds[time_name]
    ]
    if "bounds" not in ds[time_name].attrs:
        return ds
    # Handle time bounds if present
    bounds_name = ds[time_name].attrs["bounds"]
    ds[bounds_name] = np.array(
        [
            [cf.DatetimeNoLeap(t.dt.year, t.dt.month, 1) for t in ds[time_name]],
            [
                cf.DatetimeNoLeap(
                    t.dt.year if t.dt.month < 12 else (t.dt.year + 1),
                    (t.dt.month + 1) if t.dt.month < 12 else 1,
                    1,
                )
                for t in ds[time_name]
            ],
        ]
    ).T
    return ds



import numpy as np
import pandas as pd
import xarray as xr

from ilamb3 import dataset as dset


def generate_test_dset(seed: int = 1, shift: bool = False):
    rs = np.random.RandomState(seed)
    lat = [-67.5, -22.5, 22.5, 67.5]
    lon = [-135.0 + 360 * shift, -45.0 + 360 * shift, 45.0, 135.0]
    time = pd.date_range(start="2000-01-15", periods=5, freq="30D")
    ds = xr.Dataset(
        data_vars={
            "da": xr.DataArray(
                rs.rand(len(time), len(lat), len(lon)) * 1e-8,
                coords=[time, lat, lon],
                dims=["time", "lat", "lon"],
            ),
        }
    )
    ds["da"].attrs["units"] = "kg m-2 s-1"
    return ds


def generate_test_dset_with_depth(seed: int = 1):
    rs = np.random.RandomState(seed)
    lat = [-67.5, -22.5, 22.5, 67.5]
    lon = [-135.0, -45.0, 45.0, 135.0]
    depth = [0.5, 3.0, 7.5, 30.0]
    time = pd.date_range(start="2000-01-15", periods=5, freq="30D")
    ds = xr.Dataset(
        data_vars={
            "da": xr.DataArray(
                rs.rand(len(time), len(depth), len(lat), len(lon)) * 1e-8,
                coords=[time, depth, lat, lon],
                dims=["time", "depth", "lat", "lon"],
            ),
        }
    )
    ds["da"].attrs["units"] = "kg m-3 s-1"
    return ds


def generate_test_site_dset(seed: int = 1):
    rs = np.random.RandomState(seed)
    lat = xr.DataArray(data=[-67.5, -22.5, 22.5, 67.5], dims=["site"])
    lon = xr.DataArray(data=[-135.0, -45.0, 45.0, 135.0], dims=["site"])
    time = pd.date_range(start="2000-01-15", periods=5, freq="30D")
    ds = xr.Dataset(
        data_vars={
            "da": xr.DataArray(
                rs.rand(len(time), lat.size) * 1e-8,
                coords={"time": time},
                dims=["time", "site"],
            ),
        }
    )
    ds["da"] = ds["da"].assign_coords({"lat": lat, "lon": lon})
    ds["da"].attrs["units"] = "kg m-2 s-1"
    ds["da"].attrs["coordinates"] = "lat lon"
    return ds


def test_get_dim_name():
    ds = generate_test_dset()
    lon_name = dset.get_dim_name(ds, "lon")
    assert lon_name == "lon"
    try:
        dset.get_dim_name(ds, "depth")  # not in our dimensions
    except Exception as exc:
        assert isinstance(exc, KeyError)


def test_get_time_extent():
    ds = generate_test_dset()
    t0, _ = dset.get_time_extent(ds)
    assert t0 == ds["time"].isel(time=0)
    ds = ds.cf.add_bounds("time")
    t0, _ = dset.get_time_extent(ds)
    assert t0.values == pd.Timestamp("1999-12-31T00:00:00.000000000")


def test_compute_time_measures():
    ds = generate_test_dset()
    try:
        dset.compute_time_measures(ds.isel(time=slice(1)))
    except Exception as exc:
        assert isinstance(exc, ValueError)
    msr = dset.compute_time_measures(ds)
    assert np.allclose(float(msr.pint.dequantify().mean().values), 30.0)
    ds = ds.cf.add_bounds("time")
    msr = dset.compute_time_measures(ds)
    assert np.allclose(float(msr.pint.dequantify().mean().values), 30.0)


def test_coarsen_dataset():
    ds = generate_test_dset()
    ds = dset.coarsen_dataset(ds, res=90.0)
    assert np.allclose(ds["da"].values[0, 0, 0], 2.51093204e-09)


def test_integrate_time_and_space():
    ds = generate_test_dset()
    da = dset.integrate_space(dset.integrate_time(ds, "da"), "da")
    da = dset.convert(da, "Pg")
    assert np.isclose(da, 31.53474998108379)
    da = dset.integrate_time(dset.integrate_space(ds, "da", region="euro"), "da")
    da = dset.convert(da, "Pg")
    assert np.isclose(da, 8.370451774151613)


def test_integrate_space_weighted():
    ds = generate_test_dset()
    wgt = ds["da"].sum(dim="time")
    da = dset.integrate_space(ds, "da", weight=wgt)
    assert np.isclose(da.sum(), 0.30301439)


def test_integrate_depth():
    ds = generate_test_dset_with_depth()
    ds = dset.integrate_space(
        dset.integrate_time(dset.integrate_depth(ds, "da"), "da"), "da"
    )
    ds = ds.to_dataset(name="da")
    ds = dset.convert(ds, "Pg", varname="da")
    assert np.allclose(ds["da"], 1002.990371115)


def test_mean():
    ds = generate_test_dset()
    da = dset.integrate_space(dset.integrate_time(ds["da"], mean=True), "da", mean=True)
    da = dset.convert(da, "g m-2 d-1")
    assert np.isclose(da, 0.4121668497188348)


def test_std():
    ds = generate_test_dset()
    da = dset.std_time(ds["da"])
    da = da.sum()
    assert np.isclose(da, 4.343761009115869e-08)
    da = dset.std_time(ds, varname="da")
    da = da.sum()
    assert np.isclose(da, 4.343761009115869e-08)


def test_sel():
    ds = generate_test_dset()
    ds = ds.cf.add_bounds("lat")
    ds = dset.sel(ds, "lat", 0, 45)
    dlat = ds["lat_bounds"].diff(dim="bounds").values
    assert dlat.size == 1
    assert np.allclose(dlat[0, 0], 45)


def test_scale_water():
    da = xr.DataArray(1.0)
    da.attrs["units"] = "kg m-2 s-1"
    da = dset.scale_by_water_density(da, "mm d-1")
    assert np.allclose(da.pint.dequantify(), 1 / 998.2071)  # dequantify needed
    da = xr.DataArray(1.0)
    da.attrs["units"] = "mm d-1"
    da = dset.scale_by_water_density(da, "kg m-2 s-1")
    assert np.allclose(da.pint.dequantify(), 998.2071)
    da = xr.DataArray(1.0)
    da.attrs["units"] = "kg m-2"
    da = dset.scale_by_water_density(da, "mm d-1")
    assert np.allclose(da.pint.dequantify(), 1)


def test_is_spatial_or_site():
    ds = generate_test_dset()
    assert dset.is_spatial(ds)
    assert not dset.is_site(ds)
    ds = generate_test_site_dset()
    assert not dset.is_spatial(ds)
    assert dset.is_site(ds)


def test_shift_lon():
    ds = generate_test_dset(shift=True)
    ds = dset.shift_lon(ds)
    assert ds["lon"].min() < -120


def test_cell_measures():
    lats = np.linspace(-90, 90, 4)
    lons = np.linspace(-180, 180, 7)
    ds = xr.DataArray(
        np.random.rand(3, 6),
        coords={
            "lat": 0.5 * (lats[:-1] + lats[1:]),
            "lon": 0.5 * (lons[:-1] + lons[1:]),
        },
        dims=["lat", "lon"],
        name="da",
    ).to_dataset()
    ds["lat_bnds"] = (("lat", "nb"), np.array([lats[:-1], lats[1:]]).T)
    ds["lon_bnds"] = (("lon", "nb"), np.array([lons[:-1], lons[1:]]).T)
    ds["lat"].attrs["bounds"] = "lat_bnds"
    ds["lon"].attrs["bounds"] = "lon_bnds"
    assert np.allclose(dset.compute_cell_measures(ds).mean(), 2.83369151e13)



import datetime
import os
import sys

import sphinx_autosummary_accessors

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

cwd = os.getcwd()
parent = os.path.dirname(cwd)
sys.path.insert(0, parent)

# -- General configuration -----------------------------------------------------

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.autosummary",
    "sphinx_autodoc_typehints",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "numpydoc",
    "IPython.sphinxext.ipython_directive",
    "myst_nb",
    "sphinx_copybutton",
]

extlinks = {
    "issue": ("https://github.com/rubisco-sfa/ilamb3/issues%s", "GH#%s"),
    "pr": ("https://github.com/rubisco-sfa/ilamb3/pull/%s", "GH#%s"),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates", sphinx_autosummary_accessors.templates_path]

# The suffix of source filenames.
source_suffix = ".rst"

# Enable notebook execution
# https://nbsphinx.readthedocs.io/en/0.4.2/never-execute.html
# nbsphinx_execute = 'auto'
# Allow errors in all notebooks by
# nbsphinx_allow_errors = True

# Disable cell timeout
nbsphinx_timeout = -1
nb_execution_timeout = -1
nbsphinx_prolog = ""

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "ilamb3"
current_year = datetime.datetime.now().year
copyright = f"2024-{current_year}, ilamb3 Developers"
author = "ilamb Developers"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "igor"

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
css_vars = {
    "admonition-font-size": "0.9rem",
    "font-size--small": "92%",
    "font-size--small--2": "87.5%",
}

html_theme_options = dict(
    sidebar_hide_name=True,
    light_css_variables=css_vars,
    dark_css_variables=css_vars,
)

html_context = {
    "github_user": "nocollier",
    "github_repo": "ilamb3",
    "github_version": "main",
    "doc_path": "doc",
}
html_logo = "_static/logo.png"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
html_css_files = ["style.css"]

# Output file base name for HTML help builder.
htmlhelp_basename = "ilamb3doc"

# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", "ilamb3", "ilamb3 Documentation", [author], 1)]

# If true, show URL addresses after external links.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "xarray": ("https://xarray.pydata.org/en/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
}

autosummary_generate = True

autodoc_typehints = "description"
autodoc_typehints_description_target = "documented"
autodoc_default_options = {
    "members": True,
    "undoc-members": True,
    "private-members": True,
}
napoleon_use_param = True
napoleon_use_rtype = True

numpydoc_show_class_members = False
# Report warnings for all validation checks except the ones listed after "all"
numpydoc_validation_checks = {"all", "ES01", "EX01", "SA01", "SA04"}
# don't report on objects that match any of these regex
numpydoc_validation_exclude = {
    r"\.__repr__$",
}



"""Functions for use in scoring methods using regional quantiles."""

import pandas as pd
import xarray as xr

from ilamb3.exceptions import MissingRegion, NoDatabaseEntry
from ilamb3.regions import Regions


def check_quantile_database(dbase: pd.DataFrame) -> None:
    if dbase is None:
        raise ValueError("Need a quantile database")
    missing = set(dbase["region"].unique()) - set(Regions().regions)
    if missing:
        raise MissingRegion(
            "Regional quantile database uses regions with no definition in ilamb3 regions"
        )


def create_quantile_map(
    dbase: pd.DataFrame,
    quantile_variable: str,
    quantile_type: str,
    quantile_threshold: int,
) -> xr.DataArray:
    # query the database
    q = f"(quantile=={quantile_threshold})"
    q += f" & (type=='{quantile_type}')"
    q += f" & (variable=='{quantile_variable}')"
    q = dbase.query(q)
    if not len(q):
        raise NoDatabaseEntry

    # build a map
    scalar_map = Regions().region_scalars_to_map(
        {row["region"]: row["value"] for _, row in q.iterrows()}
    )
    scalar_map.attrs["units"] = q.iloc[0]["unit"]
    return scalar_map



import numpy as np
import xarray as xr

from ilamb3.analysis.runoff_sensitivity import compute_runoff_sensitivity
from ilamb3.regions import Regions
from ilamb3.tests.test_compare import generate_test_dset


def test_runoff_sensitivity():
    ilamb_regions = Regions()
    ilamb_regions.add_latlon_bounds("test", "test", [-80, 80], [-170, 170])
    ds = xr.Dataset(
        {
            "pr": 1e3 * generate_test_dset(seed=1, ntime=240, nlat=10, nlon=20)["da"],
            "tas": 273
            + 1e9 * generate_test_dset(seed=2, ntime=240, nlat=10, nlon=20)["da"],
            "mrro": generate_test_dset(seed=3, ntime=240, nlat=10, nlon=20)["da"],
        }
    )
    ds["mrro"] = (
        ds["mrro"].mean()
        + 1e2 * (ds["pr"] - ds["pr"].mean())
        + 1e-3 * (ds["tas"] - ds["tas"].mean())
    )
    df = compute_runoff_sensitivity(ds, ["test"])
    assert np.allclose(df.loc["test"]["tas Sensitivity"], 28006.878361)
    assert np.allclose(df.loc["test"]["pr Sensitivity"], 139.850275)



import tempfile
from pathlib import Path

import pytest

import ilamb3


def test_saveload():
    ilamb3.conf.reset()
    ilamb3.conf.set(
        build_dir="junkdir",
        regions=["nhsa"],
        prefer_regional_quantiles=True,
        use_uncertainty=True,
    )
    cfg_file = Path(tempfile.gettempdir()) / "tmp.cfg"
    ilamb3.conf.save(cfg_file)
    ilamb3.conf.reset()
    ilamb3.conf.load(cfg_file)
    cfg_str = f"{ilamb3.conf}"
    assert "junkdir" in cfg_str
    assert "- nhsa" in cfg_str
    ilamb3.conf.reset()


def test_context():
    ilamb3.conf.reset()
    ilamb3.conf.set(build_dir="outcontext")
    with ilamb3.conf.set(build_dir="incontext"):
        assert "incontext" in f"{ilamb3.conf}"
        assert "outcontext" not in f"{ilamb3.conf}"
    assert "incontext" not in f"{ilamb3.conf}"
    assert "outcontext" in f"{ilamb3.conf}"
    ilamb3.conf.reset()


def test_get():
    ilamb3.conf.reset()
    assert ilamb3.conf["regions"][0] is None
    assert ilamb3.conf.get("regions")[0] is None


def test_fail():
    with pytest.raises(ValueError):
        ilamb3.conf.set(regions=["not_a_region"])



"""Convenience functions which operate on datasets."""

from typing import Any, Literal

import numpy as np
import xarray as xr
from scipy.interpolate import NearestNDInterpolator

import ilamb3.regions as ilreg
from ilamb3.exceptions import NoSiteDimension, NoUncertainty


def get_dim_name(
    dset: xr.Dataset | xr.DataArray,
    dim: Literal["time", "lat", "lon", "depth", "site"],
) -> str:
    """
    Return the name of the `dim` dimension from the dataset.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The input dataset/dataarray.
    dim : str, one of {`time`, `lat`, `lon`, `depth`, `site`}
        The dimension to find in the dataset/dataarray.

    Returns
    -------
    str
        The name of the dimension.

    See Also
    --------
    get_coord_name : A variant when the coordinate is not a dimension.

    Notes
    -----
    This function is meant to handle the problem that not all data calls the dimensions
    the same things ('lat', 'Lat', 'latitude', etc). We could replace this with
    cf-xarray functionality. My concern is that we want this to work even if the
    datasets are not CF-compliant (e.g. raw model output).
    """
    dim_names = {
        "time": ["time", "TIME", "month"],
        "lat": ["lat", "latitude", "Latitude", "y", "lat_", "Lat", "LATITUDE"],
        "lon": ["lon", "longitude", "Longitude", "x", "lon_", "Lon", "LONGITUDE"],
        "depth": ["depth", "lev"],
    }
    # Assumption: the 'site' dimension is what is left over after all others are removed
    if dim == "site":
        try:
            get_dim_name(dset, "lat")
            get_dim_name(dset, "lon")
            raise NoSiteDimension("Dataset/dataarray is spatial")
        except KeyError:
            pass
        possible_names = list(
            set(dset.dims) - set([d for _, dims in dim_names.items() for d in dims])
        )
        if len(possible_names) == 1:
            return possible_names[0]
        msg = f"Ambiguity in locating a site dimension, found: {possible_names}"
        raise NoSiteDimension(msg)
    possible_names = dim_names[dim]
    dim_name = set(dset.dims).intersection(possible_names)
    if len(dim_name) != 1:
        msg = f"{dim} dimension not found: {dset.dims} "
        msg += f"not in [{','.join(possible_names)}]"
        raise KeyError(msg)
    return str(dim_name.pop())


def get_coord_name(
    dset: xr.Dataset | xr.DataArray,
    coord: Literal["lat", "lon"],
) -> str:
    """
    Return the name of the `coord` coordinate from the dataset.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The input dataset/dataarray.
    coord : str, one of {`lat`, `lon`}
        The coordinate to find in the dataset/dataarray.

    Returns
    -------
    str
        The name of the coordinate.

    See Also
    --------
    get_dim_name : A variant when the coordinate is a dimension.
    """
    coord_names = {
        "lat": ["lat", "latitude", "Latitude", "y", "lat_", "Lat", "LATITUDE"],
        "lon": ["lon", "longitude", "Longitude", "x", "lon_", "Lon", "LONGITUDE"],
    }
    possible_names = coord_names[coord]
    coord_name = set(dset.coords).intersection(possible_names)
    if len(coord_name) != 1:
        msg = f"{coord} coordinate not found: {dset.coords} "
        msg += f"not in [{','.join(possible_names)}]"
        raise KeyError(msg)
    return str(coord_name.pop())


def is_temporal(da: xr.DataArray) -> bool:
    """
    Return if the dataarray is temporal.

    Parameters
    ----------
    da : xr.DataArray
        The input dataarray.

    Returns
    -------
    bool
        True if time dimension is present, False otherwise.
    """
    try:
        get_dim_name(da, "time")
        return True
    except KeyError:
        pass
    return False


def is_spatial(da: xr.DataArray) -> bool:
    """
    Return if the dataarray is spatial.

    Parameters
    ----------
    da : xr.DataArray
        The input dataarray.

    Returns
    -------
    bool
        True if latitude and longitude dimensions are present, False otherwise.
    """
    try:
        get_dim_name(da, "lat")
        get_dim_name(da, "lon")
        return True
    except KeyError:
        pass
    # Ocean grids have 2D lat/lons sometimes stored in their coordinates
    if is_latlon2d(da):
        return True
    return False


def is_site(da: xr.DataArray) -> bool:
    """
    Return if the dataarray is a collection of sites.

    Parameters
    ----------
    da : xr.DataArray
        The input dataarray.

    Returns
    -------
    bool
        True if sites, False otherwise.
    """
    try:
        dim_lat = get_dim_name(da, "lat")
        dim_lon = get_dim_name(da, "lon")
    except KeyError:
        dim_lat = dim_lon = None
    try:
        coord_lat = get_coord_name(da, "lat")
        coord_lon = get_coord_name(da, "lon")
    except KeyError:
        return False
    if (
        dim_lat is None
        and coord_lat is not None
        and dim_lon is None
        and coord_lon is not None
    ):
        return True
    return False


def is_layered(da: xr.DataArray) -> bool:
    """
    Return if the dataarray is layered.

    Parameters
    ----------
    da : xr.DataArray
        The input dataarray.

    Returns
    -------
    bool
        True if a depth dimension is present, False otherwise.
    """
    try:
        get_dim_name(da, "depth")
        return True
    except KeyError:
        pass
    return False


def get_integer_dims(da: xr.DataArray) -> list[str]:
    """
    Return which dimensions are integers and therefore likely indices.
    """
    return [d for d in da.dims if d in da.coords and da[d].dtype.kind == "i"]


def is_latlon2d(da: xr.DataArray) -> bool:
    """
    Return if the dataarray has 2D latitudes and longitudes.
    """
    try:
        lat_coord = get_coord_name(da, "lat")
        lon_coord = get_coord_name(da, "lon")
    except KeyError:
        return False
    index_dims = get_integer_dims(da)
    if not index_dims:
        return False
    # Are the index dimensions of da also the dimensions of the the coordinates?
    if set(da[lon_coord].dims) == set(da[lat_coord].dims) == set(index_dims):
        return True
    return False


def latlon2d_to_1d(grid: xr.Dataset | xr.DataArray, da: xr.DataArray) -> xr.DataArray:
    """
    Return the da with 2D lat/lon interpolated at the grid resolution.
    """
    if not is_latlon2d(da):
        raise ValueError("Input dataarray is not a 2D lat/lon.")
    lat_dim = get_dim_name(grid, "lat")
    lon_dim = get_dim_name(grid, "lon")
    lat_coord = get_coord_name(da, "lat")
    lon_coord = get_coord_name(da, "lon")
    index_dims = get_integer_dims(da)
    da[lon_coord] = xr.where(da[lon_coord] > 180, da[lon_coord] - 360, da[lon_coord])
    # Create an interpolator that maps lat,lon --> i,j
    index_map = NearestNDInterpolator(
        np.array([da[lat_coord].values.flatten(), da[lon_coord].values.flatten()]).T,
        np.array(
            [
                (da[index_dims[0]] * xr.ones_like(da[index_dims[1]])).values.flatten(),
                (xr.ones_like(da[index_dims[0]]) * da[index_dims[1]]).values.flatten(),
            ]
        ).T,
    )
    # Then use the interpolator to find index arrays at the input regular grid.
    ids = index_map(
        (grid[lat_dim] * xr.ones_like(grid[lon_dim])).values.flatten(),
        (xr.ones_like(grid[lat_dim]) * grid[lon_dim]).values.flatten(),
    ).astype(int)
    # Create a new dataarray
    not_index_dims = [d for d in da.dims if d not in index_dims]
    coords = {c: da[c] for c in not_index_dims if c in da.coords}
    coords.update({"lat": grid["lat"], "lon": grid["lon"]})
    shp = (grid[lat_dim].size, grid[lon_dim].size)
    out = xr.DataArray(
        data=da.to_numpy()[..., ids[:, 0].reshape(shp), ids[:, 1].reshape(shp)],
        coords=coords,
        dims=not_index_dims + [lat_dim, lon_dim],
        attrs=da.attrs,
    )
    return out


def get_time_extent(
    dset: xr.Dataset | xr.DataArray, include_bounds: bool = True
) -> tuple[xr.DataArray, xr.DataArray]:
    """
    Return the time extent of the dataset/dataarray.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The input dataset.

    Returns
    -------
    tuple of xr.DataArray
        The minimum and maximum time.

    Notes
    -----
    The function will prefer the values in the 'bounds' array if present.
    """
    time_name = get_dim_name(dset, "time")
    time = dset[time_name]
    if "bounds" in time.attrs and include_bounds:
        if time.attrs["bounds"] in dset:
            time = dset[time.attrs["bounds"]]
    time.load()
    return time.min(), time.max()


def compute_time_measures(dset: xr.Dataset | xr.DataArray) -> xr.DataArray:
    """
    Return the length of each time interval.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The input dataset.

    Returns
    -------
    xr.DataArray
        The time measures, the length of the time intervals.

    Notes
    -----
    In order to integrate in time, we need the time measures. While this function is
    written for greatest flexibility, the most accurate time measures will be computed
    when a dataset is passed in where the 'bounds' on the 'time' dimension are labeled
    and part of the dataset.
    """

    def _measure1d(time, time_name):  # numpydoc ignore=GL08
        if time.size == 1:
            msg = "Cannot estimate time measures from single value without bounds"
            raise ValueError(msg)
        delt = time.diff(dim=time_name).to_numpy().astype(float) * 1e-9 / 3600 / 24
        delt = np.hstack([delt[0], delt, delt[-1]])
        msr = xr.DataArray(
            0.5 * (delt[:-1] + delt[1:]), coords=[time], dims=[time_name]
        )
        msr.attrs["units"] = "d"
        return msr

    time_name = get_dim_name(dset, "time")
    time = dset[time_name]
    timeb_name = time.attrs["bounds"] if "bounds" in time.attrs else None
    if timeb_name is None or timeb_name not in dset:
        return _measure1d(time, time_name)
    # compute from the bounds
    delt = dset[timeb_name]
    nbnd = delt.dims[-1]
    delt = delt.diff(nbnd).squeeze().compute()
    measure = delt.astype("float") * 1e-9 / 86400  # [ns] to [d]
    measure.attrs["units"] = "d"
    return measure


def compute_cell_measures(dset: xr.Dataset | xr.DataArray) -> xr.DataArray:
    """
    Return the area of each spatial cell.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The input dataset.

    Returns
    -------
    xr.DataArray
        The cell areas.

    Notes
    -----
    It would be better to get these from the model data itself, but they are not always
    provided, particularly in reference data.
    """
    earth_radius = 6.371e6  # [m]
    lat_name = get_dim_name(dset, "lat")
    lon_name = get_dim_name(dset, "lon")
    lat = dset[lat_name]
    lon = dset[lon_name]
    latb_name = lat.attrs["bounds"] if "bounds" in lat.attrs else None
    lonb_name = lon.attrs["bounds"] if "bounds" in lon.attrs else None
    # we prefer to compute your cell areas from the lat/lon bounds if they are
    # part of the dataset...
    if (
        latb_name is not None
        and latb_name in dset
        and lonb_name is not None
        and lonb_name in dset
    ):
        delx = dset[lonb_name] * np.pi / 180
        dely = np.sin(dset[latb_name] * np.pi / 180)
        other_dims = delx.dims[-1]
        delx = earth_radius * delx.diff(other_dims).squeeze()
        dely = earth_radius * dely.diff(other_dims).squeeze()  # type: ignore
        msr = dely * delx
        msr.attrs["units"] = "m2"
        return msr
    # ...and if they aren't, we assume the lat/lon we have is a cell centroid
    # and compute the area.
    lon = lon.values
    lat = lat.values
    delx = 0.5 * (lon[:-1] + lon[1:])
    dely = 0.5 * (lat[:-1] + lat[1:])
    delx = np.vstack(
        [
            np.hstack([lon[0] - 0.5 * (lon[1] - lon[0]), delx]),
            np.hstack([delx, lon[-1] + 0.5 * (lon[-1] - lon[-2])]),
        ]
    ).T
    dely = np.vstack(
        [
            np.hstack([lat[0] - 0.5 * (lat[1] - lat[0]), dely]),
            np.hstack([dely, lat[-1] + 0.5 * (lat[-1] - lat[-2])]),
        ]
    ).T
    delx = delx * np.pi / 180
    dely = np.sin(dely * np.pi / 180)
    delx = earth_radius * np.diff(delx, axis=1).squeeze()
    dely = earth_radius * np.diff(dely, axis=1).squeeze()
    delx = xr.DataArray(
        data=np.abs(delx), dims=[lon_name], coords={lon_name: dset[lon_name]}
    )
    dely = xr.DataArray(
        data=np.abs(dely), dims=[lat_name], coords={lat_name: dset[lat_name]}
    )
    msr = dely * delx
    msr.attrs["units"] = "m2"
    return msr


def coarsen_dataset(dset: xr.Dataset, res: float = 0.5) -> xr.Dataset:
    """
    Return the mass-conversing spatially coarsened dataset.

    Parameters
    ----------
    dset : xr.Dataset
        The input dataset.
    res : float, optional
        The target resolution in degrees.

    Returns
    -------
    xr.Dataset
        The coarsened dataset.

    Notes
    -----
    Coarsens the source dataset to the target resolution while conserving the
    overall integral and apply masks where all values are nan.
    """
    lat_name = get_dim_name(dset, "lat")
    lon_name = get_dim_name(dset, "lon")
    fine_per_coarse = int(
        round(res / np.abs(dset[lat_name].diff(lat_name).mean().values))  # type: ignore
    )
    # To spatially coarsen this dataset we will use the xarray 'coarsen'
    # functionality. However, if we want the area weighted sums to be the same,
    # we need to integrate over the coarse cells and then divide through by the
    # new areas. We also need to keep track of nan's to apply a mask to the
    # coarsened dataset.
    if "cell_measures" not in dset:
        dset["cell_measures"] = compute_cell_measures(dset)
    nll = (
        dset.notnull()
        .any(dim=[d for d in dset.dims if d not in [lat_name, lon_name]])
        .coarsen({"lat": fine_per_coarse, "lon": fine_per_coarse}, boundary="pad")
        .sum()  # type: ignore
        .astype(int)
    )
    dset_coarse = (
        (dset.drop_vars("cell_measures") * dset["cell_measures"])
        .coarsen({"lat": fine_per_coarse, "lon": fine_per_coarse}, boundary="pad")
        .sum()  # type: ignore
    )
    cell_measures = compute_cell_measures(dset_coarse)
    dset_coarse = dset_coarse / cell_measures
    dset_coarse["cell_measures"] = cell_measures
    dset_coarse = xr.where(nll == 0, np.nan, dset_coarse)
    return dset_coarse


def integrate_time(
    dset: xr.Dataset | xr.DataArray,
    varname: str | None = None,
    mean: bool = False,
) -> xr.DataArray:
    """
    Return the time integral or mean of the dataset.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The input dataset/dataarray.
    varname : str, optional
        The variable to integrate, must be given if a dataset is passed in.
    mean : bool, optional
        Enable to divide the integral by the integral of the measures, returning the
        mean in a functional sense.

    Returns
    -------
    integral
        The integral or mean.

    Notes
    -----
    This interface is useful in our analysis as many times we want to report the total
    of a quantity (total mass of carbon) and other times we want the mean value (e.g.
    temperature). This allows the analysis code to read the same where a flag can be
    passed to change the behavior. We could consider replacing with xarray.integrate.
    However, as of `v2023.6.0`, this does not handle the `pint` units correctly, and can
    only be done in a single dimension at a time, leaving the spatial analog to be hand
    coded. It also uses trapezoidal rule which should return the same integration, but
    could have small differences depending on how endpoints are interpretted.
    """
    time_name = get_dim_name(dset, "time")
    if isinstance(dset, xr.Dataset):
        assert varname is not None
        var = dset[varname]
        msr = (
            dset["time_measures"]
            if "time_measures" in dset
            else compute_time_measures(dset)
        )
    else:
        var = dset
        msr = compute_time_measures(dset)
    if mean:
        return var.weighted(msr.fillna(0)).mean(dim=time_name)
    var = var.pint.quantify()
    msr = msr.pint.quantify()
    out = var.weighted(msr.fillna(0)).sum(dim=time_name)
    out = out.pint.dequantify()
    return out


def accumulate_time(
    dset: xr.Dataset | xr.DataArray,
    varname: str | None = None,
) -> xr.DataArray:
    """
    Return the time accumulation of the dataset.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The input dataset/dataarray.
    varname : str, optional
        The variable to integrate, must be given if a dataset is passed in.

    Returns
    -------
    xr.DataArray
        The accumulation.
    """
    time_name = get_dim_name(dset, "time")
    if isinstance(dset, xr.Dataset):
        assert varname is not None
        var = dset[varname]
        msr = (
            dset["time_measures"]
            if "time_measures" in dset
            else compute_time_measures(dset)
        )
    else:
        var = dset
        msr = compute_time_measures(dset)
    var = (var.pint.quantify() * msr.pint.quantify()).pint.dequantify()
    var = var.cumsum(dim=time_name)
    return var


def std_time(
    dset: xr.Dataset | xr.DataArray, varname: str | None = None
) -> xr.DataArray:
    """
    Return the standard deviation of a variable in time.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The input dataset/dataarray.
    varname : str, optional
        The variable, must be given if a dataset is passed in.

    Returns
    -------
    xr.DataArray
        The weighted standard deviation.
    """
    time_name = get_dim_name(dset, "time")
    if isinstance(dset, xr.Dataset):
        var = dset[varname]
        msr = (
            dset["time_measures"]
            if "time_measures" in dset
            else compute_time_measures(dset)
        )
    else:
        var = dset
        msr = compute_time_measures(dset)
    return var.weighted(msr.fillna(0)).std(dim=time_name)


def integrate_space(
    dset: xr.DataArray | xr.Dataset,
    varname: str,
    region: None | str = None,
    mean: bool = False,
    weight: xr.DataArray | None = None,
) -> xr.DataArray:
    """
    Return the space integral or mean of the dataset.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The input dataset/dataarray.
    varname : str, optional
        The variable to integrate, must be given if a dataset is passed in.
    region : str, optional
        The region label, one of `ilamb3.Regions.regions` or `None` to indicate that the
        whole spatial domain should be used.
    mean : bool, optional
        Enable to divide the integral by the integral of the measures, returning the
        mean in a functional sense.
    weight : xr.DataArray, optional
        Optional weight for the spatial integral. Used when mass weighting.

    Returns
    -------
    xr.Dataset
        The integral or mean.

    Notes
    -----
    This interface is useful in our analysis as many times we want to report the total
    of a quantity (total mass of carbon) and other times we want the mean value (e.g.
    temperature). This allows the analysis code to read the same where a flag can be
    passed to change the behavior. We could consider replacing with xarray.integrate.
    However, as of `v2023.6.0`, this does not handle the `pint` units correctly, and can
    only be done in a single dimension at a time.
    """
    dset = dset.pint.dequantify()
    if region is not None:
        regions = ilreg.Regions()
        dset = regions.restrict_to_region(dset, region)
    space = [get_dim_name(dset, "lat"), get_dim_name(dset, "lon")]
    if not isinstance(dset, xr.Dataset):
        dset = dset.to_dataset()
    var = dset[varname]
    msr = (
        dset["cell_measures"]
        if "cell_measures" in dset
        else compute_cell_measures(dset)
    )
    if weight is not None:
        assert isinstance(weight, xr.DataArray)
        msr = (msr.pint.quantify() * weight.pint.quantify()).pint.dequantify()
    out = var.weighted(msr.fillna(0))
    if mean:
        out = out.mean(dim=space)
        out.attrs["units"] = var.attrs["units"]
    else:
        out = out.sum(dim=space)
        out.attrs["units"] = f"({var.attrs['units']})*({msr.attrs['units']})"
    return out


def sel(dset: xr.Dataset, coord: str, cmin: Any, cmax: Any) -> xr.Dataset:
    """
    Return a selection of the dataset.

    Parameters
    ----------
    dset : xr.Dataset
        The input dataset.
    coord : str
        The coordinate to slice.
    cmin :  Any
        The minimum coordinate value.
    cmax : Any
        The maximum coordinate value.

    Returns
    -------
    xr.Dataset
        The selected dataset.

    Notes
    -----
    The behavior of xarray.sel does not work for us here. We want to pick a slice of the
    dataset but where the value lies in between the coord bounds. Then we clip the min
    and max to be the limits of the slice.
    """

    def _get_interval(dset, dim, value, side):  # numpydoc ignore=GL08
        coord = dset[dim]
        if "bounds" in coord.attrs:
            if coord.attrs["bounds"] in dset:
                coord = dset[coord.attrs["bounds"]]
                ind = ((coord[:, 0] <= value) & (coord[:, 1] >= value)).to_numpy()
                ind = np.where(ind)[0]
                assert len(ind) <= 2
                assert len(ind) > 0
                if len(ind) == 2:
                    if side == "low":
                        return ind[1]
                return ind[0]
        raise NotImplementedError(f"Must have a bounds {coord}")

    dset = dset.isel(
        {
            coord: slice(
                _get_interval(dset, coord, cmin, "low"),
                _get_interval(dset, coord, cmax, "high") + 1,
            )
        }
    )
    # adjust the bounds and coord values
    bnds = dset[coord].attrs["bounds"]
    dset[bnds][0, 0] = cmin
    dset[bnds][-1, 1] = cmax
    dim = dset[coord].to_numpy()
    dim[0] = dset[bnds][0, 0].values + 0.5 * (
        dset[bnds][0, 1].values - dset[bnds][0, 0].values
    )
    dim[-1] = dset[bnds][-1, 0].values + 0.5 * (
        dset[bnds][-1, 1].values - dset[bnds][-1, 0].values
    )
    attrs = dset[coord].attrs
    dset[coord] = dim
    dset[coord].attrs = attrs
    return dset


def integrate_depth(
    dset: xr.Dataset | xr.DataArray,
    varname: str | None = None,
    mean: bool = False,
) -> xr.DataArray:
    """
    Return the depth integral or mean of the dataset.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The dataset of dataarray to integrate.
    varname : str, optional
        If dset is a dataset, the variable name to integrate.
    mean : bool, optional
        Enable to take a depth mean.

    Returns
    -------
    xr.DataArray
        The depth integral or sum.
    """
    if isinstance(dset, xr.DataArray):
        varname = dset.name
        dset = dset.to_dataset(name=varname)
    else:
        assert varname is not None
    var = dset[varname]

    # do we have a depth dimension
    try:
        depth_name = get_dim_name(dset, "depth")
    except KeyError:
        raise ValueError("Cannot integrate in depth without a depth dimension.")

    # does depth have bounds?
    if (
        "bounds" not in dset[depth_name].attrs
        or dset[depth_name].attrs["bounds"] not in dset
    ):
        dset = dset.cf.add_bounds(depth_name)

    # compute measures
    msr = dset[dset[depth_name].attrs["bounds"]]
    msr = msr.diff(dim=msr.dims[-1])
    msr.attrs["units"] = (
        dset[depth_name].attrs["units"] if "units" in dset[depth_name].attrs else "m"
    )

    # integrate
    out = var.weighted(msr.fillna(0))
    if mean:
        out = out.mean(dim=depth_name)
        out.attrs["units"] = var.attrs["units"]
    else:
        out = out.sum(dim=depth_name)
        out.attrs["units"] = f"({var.attrs['units']})*({msr.attrs['units']})"
    return out


def scale_by_water_density(da: xr.DataArray, target: str) -> xr.DataArray:
    """
    Conditionally scale the dataarray by density if needed in the conversion.

    Parameters
    ----------
    da : xr.DataArray
        The pint quantified input data array.
    target : str
        The target conversion unit as a string.

    Returns
    -------
    xr.DataArray
        The potentially water density scaled data array.

    Notes
    -----
    Most modeled hydrologic quantities tend to be output as a mass flux rate. However, a
    linear speed is often the preferred unit. For example a conversion such as `kg m-2
    s-1` to `mm d-1`. This is possible if scaled by the density of water which we do
    here conditionally if needed. Used by our `convert()` routine.
    """
    da = da.pint.quantify()
    ureg = da.pint.registry
    water_density = 998.2071 * ureg.kilogram / ureg.meter**3
    src = 1.0 * da.pint.units
    tar = ureg(target)
    if (src / tar).check("[mass] / [length]**3"):
        return da / water_density
    if (tar / src).check("[mass] / [length]**3"):
        return da * water_density
    return da


def convert(
    dset: xr.Dataset | xr.DataArray,
    unit: str,
    varname: str | None = None,
) -> xr.Dataset | xr.DataArray:
    """
    Convert the units of the dataarray.

    Parameters
    ----------
    dset : xr.Dataset or xr.DataArray
        The dataset (specify varname) or dataarray who units you wish to convert.
    unit : str
        The unit to which we will convert.
    varname : str, optional
        If dset is a dataset, give the variable name to convert.

    Returns
    -------
    xr.Dataset or xr.DataArray
        The converted dataset.
    """
    dset = dset.pint.quantify()
    if isinstance(dset, xr.DataArray):
        da = dset
    else:
        assert varname is not None
        da = dset[varname]
    da = scale_by_water_density(da, unit)
    da = da.pint.to(unit)
    if isinstance(dset, xr.DataArray):
        return da.pint.dequantify()
    dset[varname] = da
    dset = dset.pint.dequantify()
    return dset


def coarsen_annual(dset: xr.Dataset) -> xr.Dataset:
    """
    Return the dataset coarsened to annual.

    Parameters
    ----------
    dset : xr.Dataset
        The input dataset/dataarray.

    Returns
    -------
    xr.Dataset
        The coarsened dataset.
    """
    # Can't sum time objects so find them and remove
    time_name = get_dim_name(dset, "time")
    if "bounds" in dset[time_name].attrs:
        bounds = dset[time_name].attrs["bounds"]
        if bounds in dset:
            dset = dset.drop_vars(bounds)
    msr = compute_time_measures(dset).pint.dequantify()
    ann = (dset * msr).groupby(f"{time_name}.year").sum() / msr.groupby(
        f"{time_name}.year"
    ).sum()
    return ann


def shift_lon(dset: xr.Dataset) -> xr.Dataset:
    """
    Return the dataset with longitudes shifted to [-180,180].

    Parameters
    ----------
    dset : xr.Dataset
        The input dataset.

    Returns
    -------
    xr.Dataset
        The longitude-shifted dataset.
    """
    lon_name = get_coord_name(dset, "lon")
    if (dset[lon_name].min() >= 0) * (dset[lon_name].max() <= 360):
        dset[lon_name] = (dset[lon_name] + 180) % 360 - 180
        if "bounds" in dset[lon_name].attrs and dset[lon_name].attrs["bounds"] in dset:
            dset[dset[lon_name].attrs["bounds"]] = (
                dset[dset[lon_name].attrs["bounds"]] + 180
            ) % 360 - 180
        dset = dset.sortby(lon_name)
    return dset


def get_scalar_uncertainty(ds: xr.Dataset, varname: str) -> xr.DataArray:
    """
    Get a scalar uncertainty from the variable.

    Note
    ----
    Uncertainty is indicated either by the presence of the `ancillary_variables`
    attribute or the `bounds` attribute. In the case of the latter, we will
    return the harmonic mean as a scalar measure of uncertainty.
    """
    var = ds[varname]
    da = None
    if "ancillary_variables" in var.attrs and var.attrs["ancillary_variables"] in ds:
        da = ds[var.attrs["ancillary_variables"]]
    if "bounds" in var.attrs and var.attrs["bounds"] in ds:
        da = ds[var.attrs["bounds"]]
        bnd_dim = set(da.dims) - set(var.dims)
        if len(bnd_dim) > 1:
            raise ValueError(
                f"Ambiguity in determinging the `bounds` dimension, found: {bnd_dim}"
            )
        bnd_dim = list(bnd_dim)[0]
        da = np.sqrt(
            (var - da.isel({bnd_dim: 0})) ** 2 + (da.isel({bnd_dim: 1}) - var) ** 2
        )
    if da is None:
        raise NoUncertainty()
    da.attrs["units"] = var.attrs["units"]
    return da


def fix_missing_bounds_attrs(ds: xr.Dataset) -> xr.Dataset:
    """
    Add a `bounds` attribute if variable appears to exist.
    """

    def _fix(ds: xr.Dataset, dim_name: str) -> xr.Dataset:
        dim = ds[dim_name]
        if "bounds" in dim.attrs:
            if dim.attrs["bounds"] in ds:
                return ds  # all good, just return
            else:
                ds[dim_name].attrs.pop("bounds")
                return ds
        # is there a likely bounds variable and the attr is missing?
        for possible_bnds in [f"{dim_name}_bnds", f"{dim_name}_bounds"]:
            if possible_bnds in ds:
                ds[dim_name].attrs["bounds"] = possible_bnds
        return ds

    for d in ds.dims:
        ds = _fix(ds, d)
    return ds


def get_mean_time_frequency(ds: xr.Dataset) -> float:
    """
    Return the mean time frequency of the dataset.
    """
    tm = compute_time_measures(ds)
    return float(tm.mean())


def compute_monthly_mean(ds: xr.Dataset) -> xr.Dataset:
    """
    Return the monthly mean of the input dataset in a `noleap` calendar.
    """
    if not is_temporal(ds):
        raise ValueError("Input dataset has no temporal dimension.")
    dt = get_mean_time_frequency(ds)
    if dt > 31:
        raise ValueError(f"Input dataset is already coarser than monthly {dt=}.")
    ds = ds.resample(time=xr.groupers.TimeResampler("MS")).mean()
    return ds



import numpy as np
import pandas as pd
import pytest

from ilamb3.analysis import bias_analysis
from ilamb3.regions import Regions
from ilamb3.tests.test_compare import generate_test_dset
from ilamb3.tests.test_dataset import generate_test_site_dset


def gen_quantile_dbase(seed=1):
    rs = np.random.RandomState(seed)
    df = []
    for r in Regions().regions:
        for th in [70, 80]:
            df.append(
                {
                    "variable": "da",
                    "region": r,
                    "quantile": th,
                    "type": "bias",
                    "value": rs.rand(1)[0] * 1e-9,
                    "unit": "kg m-2 s-1",
                }
            )
    return pd.DataFrame(df)


@pytest.mark.parametrize(
    "use_uncertainty,mass_weighting,score",
    [
        (True, False, 0.5037343625713414),
        (False, False, 0.49809129117395395),
        (True, True, 0.6211524133325482),
        (False, True, 0.6162697692652096),
    ],
)
def test_bias_collier2018(use_uncertainty: bool, mass_weighting: bool, score: float):
    grid = dict(nlat=10, nlon=20)
    ref = generate_test_dset(**grid)
    ref["da_bnds"] = generate_test_dset(seed=2, **grid)["da"] * 1e-2
    ref["da"].attrs["bounds"] = "da_bnds"
    com = generate_test_dset(seed=3, **grid)
    analysis = bias_analysis(
        "da",
        method="Collier2018",
        use_uncertainty=use_uncertainty,
        mass_weighting=mass_weighting,
    )
    df, _, _ = analysis(ref, com)
    df = df[df["type"] == "score"]
    assert len(df) == 1
    assert np.allclose(df.iloc[0].value, score)


def test_bias_site_collier2018():
    ref = generate_test_site_dset().mean(dim="time")
    com = generate_test_dset(nlat=10, nlon=20)
    analysis = bias_analysis("da")
    assert set(["da"]) == set(analysis.required_variables())
    df, _, _ = analysis(ref, com)
    df = df[df["type"] == "score"]
    assert len(df) == 1
    assert np.allclose(df.iloc[0].value, 0.665078089592162)


@pytest.mark.parametrize(
    "use_uncertainty,quantile_threshold,score",
    [
        (True, 80, 0.038101086238025224),
        (False, 80, 0.03181825609556247),
        (False, 70, 0.01916371991867042),
    ],
)
def test_bias_regionalquantiles(
    use_uncertainty: bool, quantile_threshold: bool, score: float
):
    grid = dict(nlat=10, nlon=20)
    ref = generate_test_dset(seed=1, **grid)
    ref["da_bnds"] = generate_test_dset(**grid)["da"] * 1e-2
    ref["da"].attrs["bounds"] = "da_bnds"
    com = generate_test_dset(seed=2, **grid)
    analysis = bias_analysis(
        "da",
        method="RegionalQuantiles",
        use_uncertainty=use_uncertainty,
        quantile_database=gen_quantile_dbase(),
        quantile_threshold=quantile_threshold,
    )
    df, _, _ = analysis(ref, com)
    df = df[df["type"] == "score"]
    print(df.iloc[0].value)
    assert len(df) == 1
    assert np.allclose(df.iloc[0].value, score)



"""Functions for synthesizing ilamb3 output."""

import importlib
import json
from pathlib import Path
from typing import Any

import pandas as pd
from jinja2 import Template

import ilamb3.regions as ilr
from ilamb3.analysis import add_overall_score


def dataframe_to_cmec(df: pd.DataFrame) -> dict[str, Any]:
    """
    Build a CMEC boundle from information in the dataframe.
    """
    df = df.drop(columns=["analysis", "type", "units"])
    df["h2"] = df["section"] + "::" + df["variable"]
    df["h3"] = df["h2"] + "!!" + df["dataset"]
    ilamb_regions = ilr.Regions()
    cmec_regions = {
        r: {
            "LongName": "None" if r == "None" else ilamb_regions.get_name(r),
            "Description": (
                "Reference data extents" if r == "None" else ilamb_regions.get_name(r)
            ),
            "Generator": "N/A" if r == "None" else ilamb_regions.get_source(r),
        }
        for r in df["region"].unique()
    }
    cmec_models = {
        m: {"Description": m, "Source": m}
        for m in df["source"].unique()
        if m != "Reference"
    }
    cmec_h1 = {
        name: {
            "name": name,
            "Abstract": "composite score",
            "URI": [
                "https://www.osti.gov/biblio/1330803",
                "https://doi.org/10.1029/2018MS001354",
            ],
            "Contact": "forrest AT climatemodeling.org",
        }
        for name in df["section"].unique()
    }
    cmec_h2 = {
        name: {
            "name": name,
            "Abstract": "composite score",
            "URI": [
                "https://www.osti.gov/biblio/1330803",
                "https://doi.org/10.1029/2018MS001354",
            ],
            "Contact": "forrest AT climatemodeling.org",
        }
        for name in df["h2"].unique()
    }
    cmec_h3 = {
        name: {
            "name": name,
            "Abstract": "benchmark score",
            "URI": [
                "https://www.osti.gov/biblio/1330803",
                "https://doi.org/10.1029/2018MS001354",
            ],
            "Contact": "forrest AT climatemodeling.org",
        }
        for name in df["h3"].unique()
    }
    cmec_statistics = {
        "indices": [s.replace(" [1]", "") for s in df["name"].unique()],
        "short_names": [s.replace(" [1]", "") for s in df["name"].unique()],
    }
    cmec_results = {}
    for region in df["region"].unique():
        cmec_results[region] = {}
        for model in df["source"].unique():
            cmec_results[region][model] = {}

            for sec in ["section", "h2", "h3"]:
                for name in df[sec].unique():
                    q = df[
                        (df["region"] == region)
                        & (df["source"] == model)
                        & (df[sec] == name)
                    ]
                    if not len(q):
                        continue
                    q.groupby("name").mean(numeric_only=True)
                    q = (
                        q[["name", "dataset", "value"]]
                        .pivot(columns="name", index="dataset")
                        .mean()
                    )
                    q.index = [i.replace(" [1]", "") for i in q.index.levels[1]]
                    cmec_results[region][model][name] = q.to_dict()

    bundle = {
        "SCHEMA": {"name": "CMEC", "version": "v1", "package": "ILAMB"},
        "DIMENSIONS": {
            "json_structure": ["region", "model", "metric", "statistic"],
            "dimensions": {
                "region": cmec_regions,
                "model": cmec_models,
                "metric": cmec_h1 | cmec_h2 | cmec_h3,
                "statistic": cmec_statistics,
            },
        },
        "RESULTS": cmec_results,
    }
    return bundle


def generate_dashboard_page(
    output_path: Path, page_title: str = "ILAMB Results"
) -> None:
    """
    Create the required CMEC assets for a Unified Dashboard page.
    """
    df = build_global_dataframe(output_path)
    bundle = dataframe_to_cmec(df)
    with open(output_path / "scalar_database.json", "w") as out:
        out.write(json.dumps(bundle))
    with open(output_path / "_lmtUDConfig.json", "w") as out:
        out.write(
            json.dumps(
                {
                    "udcJsonLoc": "scalar_database.json",
                    "udcDimSets": {
                        "x_dim": "model",
                        "y_dim": "metric",
                        "fxdim": {
                            "region": df["region"].value_counts().idxmax(),
                            "statistic": "Overall Score",
                        },
                    },
                    "udcScreenHeight": 0,
                    "udcCellValue": 1,
                    "udcNormType": "standarized",
                    "udcNormAxis": "row",
                    "logofile": "None",
                }
            )
        )
    template = importlib.resources.open_text(
        "ilamb3.templates", "unified_dashboard.html"
    ).read()
    html = Template(template).render({"page_title": page_title})
    with open(output_path / "index.html", "w") as out:
        out.write(html)


def _load_local_csvs(csv_files: list[Path]) -> pd.DataFrame:
    df = pd.concat(
        [pd.read_csv(str(csv_file)) for csv_file in csv_files]
    ).drop_duplicates(subset=["source", "region", "analysis", "name"])
    df["region"] = df["region"].astype(str).str.replace("nan", "None")
    df = add_overall_score(df)
    return df


def build_global_dataframe(root: Path):
    dfs = []
    for parent, _, files in root.walk():
        # the dashboard needs html files to be {DATASET}.html
        for html_file in [parent / f for f in files if f.endswith(".html")]:
            html_file.rename(html_file.parent / f"{html_file.parent.name}.html")

        csv_files = [parent / f for f in files if f.endswith(".csv")]
        if not csv_files:
            continue
        df = _load_local_csvs(csv_files)
        df = df[df.type == "score"]

        # if at this point the depth is not 3, this function will not work
        heading = parent.relative_to(root)
        if len(heading.parents) != 3:
            raise ValueError("This function requires analysis runs of depth level = 3")
        heading = str(heading).split("/")

        # add metadata to this part of the dataframe
        df["section"] = heading[0]
        df["variable"] = heading[1]
        df["dataset"] = heading[2]
        dfs.append(df)
    dfs = pd.concat(dfs)
    return dfs



"""
An ILAMB transform for computing the accumulated ocean heat content.
"""

import cftime as cf
import xarray as xr

import ilamb3
import ilamb3.dataset as dset
from ilamb3.transform.base import ILAMBTransform


class ocean_heat_content(ILAMBTransform):
    """
    An ILAMB transform for computing the accumulated ocean heat content.

    Parameters
    ----------
    reference_year : int, optional
        The year from which to compute the ocean heat content. Defaults to the
        first year of the input dataset.
    """

    def __init__(self, reference_year: int | None = None):
        self.reference_year = reference_year

    def required_variables(self) -> list[str]:
        """
        Return the required variables for the transform.
        """
        return ["thetao", "volcello"]

    def __call__(self, ds: xr.Dataset) -> xr.Dataset:
        """
        Compute the ocean heat content in [ZJ].
        """
        if "ohc" in ds:
            return ds
        if not set(self.required_variables()).issubset(ds):
            return ds
        time_name = dset.get_dim_name(ds, "time")
        if self.reference_year is None:
            self.reference_year = int(ds[time_name].min().dt.year.values)
        ds = ds.sel({time_name: slice(f"{self.reference_year}-01", None)})
        ds = ds.pint.quantify()
        Cp = 3991.868 * ilamb3.units.J / ilamb3.units.kg / ilamb3.units.K
        rho = 1026.0 * ilamb3.units.kg / ilamb3.units.m**3
        ohc = (Cp * rho * ds["thetao"] * ds["volcello"]).sum(dim=ds["volcello"].dims)
        ohc = ohc - ohc.isel({time_name: 0})
        ohc = ohc.pint.to("ZJ")
        ohc.load()
        ohc.name = "ohc"
        ohc.attrs = {
            "standard_name": "Ocean Heat Content",
            "comment": f"Relative to {self.reference_year}",
        }
        ds = ds.drop_vars(["thetao", "volcello"], errors="ignore")
        ds["ohc"] = ohc
        ds = ds.pint.dequantify()
        ds = ds.groupby(f"{time_name}.year").mean()
        ds = ds.rename(year="time")
        ds["time"] = [cf.DatetimeNoLeap(y, 6, 1) for y in ds["time"]]
        return ds



"""
An abstract class for implementing transform functions used in ILAMB.
"""

from abc import ABC, abstractmethod
from typing import Any

import xarray as xr


class ILAMBTransform(ABC):
    """
    The ILAMB transform base class.

    A transform is just a function that takes in and returns a dataset.
    """

    @abstractmethod
    def __init__(self, **kwargs: Any):
        """
        Initialize the transform with any keyword arguments.
        """
        raise NotImplementedError()

    @abstractmethod
    def required_variables(self) -> list[str]:
        """
        Return the variables that this transform needs to have in the dataset.
        """
        raise NotImplementedError()

    @abstractmethod
    def __call__(self, ds: xr.Dataset) -> xr.Dataset:
        """
        The body of the transform function.
        """
        raise NotImplementedError()



"""
The ILAMB relationship (variable to variable) methodology.

There is a lot of data to manage for relationships. So we build a relationship class and
then later create a ILAMBAnalysis which uses it.
"""

from dataclasses import dataclass, field
from typing import Any, Union

import numpy as np
import pandas as pd
import xarray as xr

import ilamb3.plot as plt
from ilamb3 import compare as cmp
from ilamb3 import dataset as dset
from ilamb3.analysis.base import ILAMBAnalysis
from ilamb3.regions import Regions


@dataclass
class Relationship:
    """
    A class for developing and comparing relationships from gridded data.
    """

    dep: xr.DataArray
    ind: xr.DataArray
    color: xr.DataArray = None
    dep_log: bool = False
    ind_log: bool = False
    dep_var: str = "dep"
    ind_var: str = "ind"
    dep_label: str = "dep"
    ind_label: str = "ind"
    order: int = 1
    _dep_limits: list[float] = field(init=False, default_factory=lambda: None)
    _ind_limits: list[float] = field(init=False, default_factory=lambda: None)
    _dist2d: np.ndarray = field(init=False, default_factory=lambda: None)
    _ind_edges: np.ndarray = field(init=False, default_factory=lambda: None)
    _dep_edges: np.ndarray = field(init=False, default_factory=lambda: None)
    _response_mean: np.ndarray = field(init=False, default_factory=lambda: None)
    _response_std: np.ndarray = field(init=False, default_factory=lambda: None)

    def __post_init__(self):
        """After initialization, perform checks on input data."""
        # check input dataarrays for compatibility
        assert isinstance(self.dep, xr.DataArray)
        assert isinstance(self.ind, xr.DataArray)
        self.dep = self.dep.sortby(list(self.dep.sizes.keys()))
        self.ind = self.ind.sortby(list(self.ind.sizes.keys()))
        self.dep, self.ind = xr.align(self.dep, self.ind, join="exact")
        if self.color is not None:
            assert isinstance(self.color, xr.DataArray)
            self.color = self.color.sortby(list(self.color.sizes.keys()))
            self.dep, self.ind, self.color = xr.align(
                self.dep, self.ind, self.color, join="exact"
            )

        # only consider where both are valid and finite
        keep = self.dep.notnull() * self.ind.notnull()
        keep *= np.isfinite(self.dep)
        keep *= np.isfinite(self.ind)
        self.dep = xr.where(keep, self.dep, np.nan)
        self.ind = xr.where(keep, self.ind, np.nan)
        if self.dep_log:
            assert self.dep.min() > 0
        if self.ind_log:
            assert self.ind.min() > 0

    def compute_limits(
        self, rel: Union["Relationship", None] = None
    ) -> Union["Relationship", None]:
        """
        Compute the limits of the dependent and independent variables.

        Parameters
        ----------
        rel : Relationship, optional
            An optional and additional relationship who limits you also would like to
            include in the check.

        Returns
        -------
        Relationship or None
            If a relationship was passed in, this will return it also with its limits
            set.
        """

        def _singlelimit(var, limit=None):  # numpydoc ignore=GL08
            lim = [var.min(), var.max()]
            delta = 1e-8 * (lim[1] - lim[0])
            lim[0] -= delta
            lim[1] += delta
            if limit is None:
                limit = lim
            else:
                limit[0] = min(limit[0], lim[0])
                limit[1] = max(limit[1], lim[1])
            return limit

        dep_lim = _singlelimit(self.dep)
        ind_lim = _singlelimit(self.ind)
        if rel is not None:
            dep_lim = _singlelimit(self.dep, limit=dep_lim)
            ind_lim = _singlelimit(self.ind, limit=ind_lim)
            rel._dep_limits = dep_lim
            rel._ind_limits = ind_lim
        self._dep_limits = dep_lim
        self._ind_limits = ind_lim
        return rel

    def build_response(self, nbin: int = 25, eps: float = 3e-3):
        """
        Create the 2D distribution and functional response.

        Parameters
        ----------
        nbin : int
            The number of bins to use in both dimensions.
        eps : float
            The fraction of points required for a bin in the
            independent variable be included in the funcitonal responses.
        """
        # if no limits have been created, make them now
        if self._dep_limits is None or self._ind_limits is None:
            self.compute_limits(None)

        # compute the 2d distribution
        ind = np.ma.masked_invalid(self.ind.values).compressed()
        dep = np.ma.masked_invalid(self.dep.values).compressed()
        xedges = nbin
        yedges = nbin
        if self.ind_log:
            xedges = 10 ** np.linspace(
                np.log10(self.ind_limits[0]), np.log10(self.ind_limits[1]), nbin + 1
            )
        if self.dep_log:
            yedges = 10 ** np.linspace(
                np.log10(self.dep_limits[0]), np.log10(self.dep_limits[1]), nbin + 1
            )
        dist, xedges, yedges = np.histogram2d(
            ind,
            dep,
            bins=[xedges, yedges],
            range=[
                [v.values for v in self._ind_limits],
                [v.values for v in self._dep_limits],
            ],
        )
        dist = np.ma.masked_values(dist.T, 0).astype(float)
        dist /= dist.sum()
        self._dist2d = dist
        self._ind_edges = xedges
        self._dep_edges = yedges

        # compute a binned functional response
        which_bin = np.digitize(ind, xedges).clip(1, xedges.size - 1) - 1
        mean = np.ma.zeros(xedges.size - 1)
        std = np.ma.zeros(xedges.size - 1)
        cnt = np.ma.zeros(xedges.size - 1)
        with np.errstate(under="ignore"):
            for i in range(mean.size):
                depi = dep[which_bin == i]
                cnt[i] = depi.size
                if cnt[i] == 0:  # will get masked out later
                    mean[i] = 0
                    std[i] = 0
                else:
                    if self.dep_log:
                        depi = np.log10(depi)
                        mean[i] = 10 ** depi.mean()
                        std[i] = 10 ** depi.std()
                    else:
                        mean[i] = depi.mean()
                        std[i] = depi.std()
            mean = np.ma.masked_array(mean, mask=(cnt / cnt.sum()) < eps)
            std = np.ma.masked_array(std, mask=(cnt / cnt.sum()) < eps)
        self._response_mean = mean
        self._response_std = std

    def score_response(self, rel: "Relationship") -> float:
        """
        Score the functional response of the relationships.

        Parameters
        ----------
        rel : Relationship
            The other relationship to compare.

        Returns
        -------
        float
            The response score.
        """
        rel_error = np.linalg.norm(
            self._response_mean - rel._response_mean
        ) / np.linalg.norm(self._response_mean)
        score = np.exp(-rel_error)
        return score

    def to_dataset(self) -> xr.Dataset:
        """
        Convert internal relationship representation to a dataset.
        """
        ds = xr.Dataset(
            data_vars={
                f"distribution-{self.ind_var}": (
                    [self.dep_var, self.ind_var],
                    self._dist2d,
                ),
                f"response-{self.ind_var}": ([self.ind_var], self._response_mean),
                f"response-{self.ind_var}-variability": (
                    [self.ind_var],
                    self._response_std,
                ),
            },
            coords={
                self.ind_var: (
                    self.ind_var,
                    0.5 * (self._ind_edges[1:] + self._ind_edges[:-1]),
                ),
                self.dep_var: (
                    self.dep_var,
                    0.5 * (self._dep_edges[1:] + self._dep_edges[:-1]),
                ),
            },
        )
        ds[f"response-{self.ind_var}"].attrs = {
            "ancillary_variables": f"response-{self.ind_var}-variability"
        }
        ds[self.ind_var].attrs = {"standard_name": self.ind_label}
        ds[self.dep_var].attrs = {"standard_name": self.dep_label}
        return ds


class relationship_analysis(ILAMBAnalysis):
    """
    The ILAMB relationship methodology.

    Parameters
    ----------
    dep_variable : str
        The name of the dependent variable to be used in this analysis.
    ind_variable : str
        The name of the independent variable to be used in this analysis.
    regions: list[str | None] = [None],
        The regions overwhich to perform the analysis.

    Methods
    -------
    required_variables
        What variables are required.
    __call__
        The method
    """

    def __init__(
        self,
        dep_variable: str,
        ind_variable: str,
        regions: list[str | None] = [None],
        **kwargs: Any,
    ):  # numpydoc ignore=GL08
        self.dep_variable = dep_variable
        self.ind_variable = ind_variable
        self.regions = regions

    def required_variables(self) -> list[str]:
        """
        Return the list of variables required for this analysis.

        Returns
        -------
        list
            The variable names used in this analysis.
        """
        return [self.dep_variable, self.ind_variable]

    def __call__(
        self, ref: xr.Dataset, com: xr.Dataset
    ) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
        """
        Apply the ILAMB relationship methodology on the given datasets.

        Parameters
        ----------
        ref : xr.Dataset
            The reference dataset.
        com : xr.Dataset
            The comparison dataset.

        Returns
        -------
        pd.DataFrame
            A dataframe with scalar and score information from the comparison.
        xr.Dataset
            A dataset containing reference grided information from the comparison.
        xr.Dataset
            A dataset containing comparison grided information from the comparison.
        """

        def _regionify(ds: xr.Dataset, region: str | None):
            """Postpend the region name to the variables/dimenions"""
            ds = ds.rename({v: f"{v}_{region}" for v in list(ds.variables)})
            for key, da in ds.items():
                if "ancillary_variables" in da.attrs:
                    ds[key].attrs[
                        "ancillary_variables"
                    ] = f"{da.attrs['ancillary_variables']}_{region}"
            return ds

        # Initialize and make comparable
        analysis_name = f"Relationship {self.ind_variable}"
        var_ind = self.ind_variable
        var_dep = self.dep_variable
        for var in self.required_variables():
            ref, com = cmp.make_comparable(ref, com, var)
            if dset.is_temporal(ref[var]):
                ref[var] = dset.integrate_time(ref[var], mean=True)
            if dset.is_temporal(com[var]):
                com[var] = dset.integrate_time(com[var], mean=True)

        # Create and score relationships per region
        dfs = []
        ds_ref = []
        ds_com = []
        ilamb_regions = Regions()
        for region in self.regions:
            refr = ilamb_regions.restrict_to_region(ref, region)
            comr = ilamb_regions.restrict_to_region(com, region)
            rel_ref = Relationship(
                refr[var_dep],
                refr[var_ind],
                dep_var=var_dep,
                ind_var=var_ind,
                dep_label=var_dep,
                ind_label=var_ind,
            )
            rel_com = Relationship(
                comr[var_dep],
                comr[var_ind],
                dep_var=var_dep,
                ind_var=var_ind,
                dep_label=var_dep,
                ind_label=var_ind,
            )
            rel_com = rel_ref.compute_limits(rel_com)
            rel_ref.build_response()
            rel_com.build_response()
            score = rel_ref.score_response(rel_com)
            ds_ref.append(_regionify(rel_ref.to_dataset(), region))
            ds_com.append(_regionify(rel_com.to_dataset(), region))
            dfs.append(
                [
                    "Comparison",
                    str(region),
                    analysis_name,
                    f"Relationship Score {var_ind}",
                    "score",
                    "1",
                    score,
                ]
            )

        # Conversions and output
        dfs = pd.DataFrame(
            dfs,
            columns=[
                "source",
                "region",
                "analysis",
                "name",
                "type",
                "units",
                "value",
            ],
        )
        ds_ref = xr.merge(ds_ref)
        ds_com = xr.merge(ds_com)
        return dfs, ds_ref, ds_com

    def plots(
        self,
        df: pd.DataFrame,
        ref: xr.Dataset,
        com: dict[str, xr.Dataset],
    ) -> pd.DataFrame:
        # Some initialization
        com["Reference"] = ref

        # Build up a dataframe of matplotlib axes, first the distribution plots
        axs = [
            {
                "name": f"distribution-{self.ind_variable}",
                "title": f"{self.dep_variable} vs. {self.ind_variable}",
                "region": region,
                "source": source,
                "axis": (
                    plt.plot_distribution(
                        ds[f"distribution-{self.ind_variable}_{region}"],
                        title=f"{source} {self.dep_variable} vs. {self.ind_variable}",
                    )
                    if f"distribution-{self.ind_variable}_{region}" in ds
                    else pd.NA
                ),
            }
            for region in self.regions
            for source, ds in com.items()
        ]
        com.pop("Reference")
        axs += [
            {
                "name": f"response-{self.ind_variable}",
                "title": f"{self.dep_variable} vs. {self.ind_variable}",
                "region": region,
                "source": source,
                "axis": (
                    plt.plot_response(
                        ref[f"response-{self.ind_variable}_{region}"],
                        ref[f"response-{self.ind_variable}-variability_{region}"],
                        ds[f"response-{self.ind_variable}_{region}"],
                        ds[f"response-{self.ind_variable}-variability_{region}"],
                        source,
                        title=f"{source} {self.dep_variable} vs. {self.ind_variable}",
                    )
                    if f"response-{self.ind_variable}_{region}" in ds
                    else pd.NA
                ),
            }
            for region in self.regions
            for source, ds in com.items()
        ]
        axs = pd.DataFrame(axs).dropna(subset=["axis"])

        return axs



import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from ilamb3.analysis import relationship_analysis
from ilamb3.tests.test_compare import generate_test_dset


def test_relationship():
    ds = xr.merge(
        [
            generate_test_dset(nlat=2, nlon=3).rename_vars({"da": "gpp"})["gpp"],
            generate_test_dset(seed=2, nlat=2, nlon=3).rename_vars({"da": "tas"})[
                "tas"
            ],
        ]
    )
    analysis = relationship_analysis("gpp", "tas")
    assert set(["gpp", "tas"]) == set(analysis.required_variables())
    ds_com = {}
    df, ds_ref, ds_com["Comparison"] = analysis(ds, ds)
    assert len(df) == 1
    assert np.allclose(df.loc[0, "value"], 1.0)
    dfp = analysis.plots(df, ds_ref, ds_com)
    assert len(dfp) == 3
    plt.close()



"""
The ILAMB bias methodology.

See Also
--------
ILAMBAnalysis : The abstract base class from which this derives.
"""

import warnings
from typing import Any, Literal

import numpy as np
import pandas as pd
import xarray as xr

import ilamb3.plot as plt
import ilamb3.regions as ilr
from ilamb3 import compare as cmp
from ilamb3 import dataset as dset
from ilamb3.analysis.base import ILAMBAnalysis
from ilamb3.analysis.quantiles import check_quantile_database, create_quantile_map
from ilamb3.exceptions import NoDatabaseEntry


class bias_analysis(ILAMBAnalysis):
    """
    The ILAMB bias methodology.

    Parameters
    ----------
    required_variable : str
        The name of the variable to be used in this analysis.
    variable_cmap : str
        The colormap to use in plots of the comparison variable, optional.
    method : str
        The name of the scoring methodology to use, either `Collier2018` or
        `RegionalQuantiles`.
    regions : list
        A list of region labels over which to apply the analysis.
    use_uncertainty : bool
        Enable to utilize uncertainty information from the reference product if
        present.
    spatial_sum : bool
        Enable to report a spatial sum in the period mean as opposed to a
        spatial mean. This is often preferred in carbon variables where the
        total global carbon is of interest.
    mass_weighting : bool
        Enable to weight the score map integrals by the temporal mean of the
        reference dataset.
    quantile_dbase : pd.DataFrame
        If using `method='RegionalQuantiles'`, the dataframe containing the
        regional quantiles to be used to score the datasets.
    quantile_threshold : int
        If using `method='RegionalQuantiles'`, the threshold values to use from
        the `quantile_dbase`.

    Methods
    -------
    required_variables
        What variables are required.
    __call__
        The method
    """

    def __init__(
        self,
        required_variable: str,
        variable_cmap: str = "viridis",
        method: Literal["Collier2018", "RegionalQuantiles"] = "Collier2018",
        regions: list[str | None] = [None],
        use_uncertainty: bool = True,
        spatial_sum: bool = False,
        mass_weighting: bool = False,
        quantile_database: pd.DataFrame | None = None,
        quantile_threshold: int = 70,
        **kwargs: Any,  # this is so we can pass extra arguments without failure
    ):
        self.req_variable = required_variable
        self.cmap = variable_cmap
        self.method = method
        self.regions = regions
        self.use_uncertainty = use_uncertainty
        self.spatial_sum = spatial_sum
        self.mass_weighting = mass_weighting
        self.quantile_database = quantile_database
        self.quantile_threshold = quantile_threshold
        self.kwargs = kwargs

    def required_variables(self) -> list[str]:
        """
        Return the list of variables required for this analysis.

        Returns
        -------
        list
            The variable names used in this analysis.
        """
        return [self.req_variable]

    def __call__(
        self,
        ref: xr.Dataset,
        com: xr.Dataset,
    ) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
        """
        Apply the ILAMB bias methodology on the given datasets.

        Parameters
        ----------
        ref : xr.Dataset
            The reference dataset.
        com : xr.Dataset
            The comparison dataset.

        Returns
        -------
        pd.DataFrame
            A dataframe with scalar and score information from the comparison.
        xr.Dataset
            A dataset containing reference grided information from the comparison.
        xr.Dataset
            A dataset containing comparison grided information from the comparison.
        """
        # Initialize
        analysis_name = "Bias"
        varname = self.req_variable
        if self.use_uncertainty and "bounds" not in ref[varname].attrs:
            self.use_uncertainty = False

        # Checks on the quantile database if used
        if self.method == "RegionalQuantiles":
            check_quantile_database(self.quantile_database)
            try:
                quantile_map = create_quantile_map(
                    self.quantile_database, varname, "bias", self.quantile_threshold
                )
                dset.convert(quantile_map, ref[varname].attrs["units"])
            except NoDatabaseEntry:
                # fallback if the variable/type/quantile is not in the database
                self.method = "Collier2018"

        # Never mass weight if regional quantiles are used
        if self.method == "RegionalQuantiles":
            self.mass_weighting = False

        # Make the variables comparable and force loading into memory
        ref, com = cmp.make_comparable(ref, com, varname)

        # Temporal means across the time period
        ref_mean = (
            dset.integrate_time(ref, varname, mean=True)
            if dset.is_temporal(ref[varname])
            else ref[varname]
        )
        com_mean = (
            dset.integrate_time(com, varname, mean=True)
            if dset.is_temporal(com[varname])
            else com[varname]
        )

        # Get the reference data uncertainty
        uncert = xr.zeros_like(ref_mean)
        uncert.attrs["units"] = ref[varname].attrs["units"]
        if self.use_uncertainty:
            uncert = ref[ref[varname].attrs["bounds"]]
            uncert.attrs["units"] = ref[varname].attrs["units"]
            uncert = (
                dset.integrate_time(uncert, mean=True)
                if dset.is_temporal(uncert)
                else uncert
            )

        # If temporal information is available, we normalize the error by the
        # standard deviation of the reference. If not, we revert to the traditional
        # definition of relative error.
        norm = ref_mean
        if dset.is_temporal(ref):
            time_dim = dset.get_dim_name(ref, "time")
            if ref[time_dim].size > 1:
                norm = dset.std_time(ref, varname)

        # Nest the grids for comparison, we postpend composite grid variables with "_"
        if dset.is_spatial(ref) and dset.is_spatial(com):
            ref_, com_, norm_, uncert_ = cmp.nest_spatial_grids(
                ref_mean, com_mean, norm, uncert
            )
        elif dset.is_site(ref) and dset.is_site(com):
            ref_ = ref_mean
            com_ = com_mean
            norm_ = norm
            uncert_ = uncert
        else:
            raise ValueError("Reference and comparison not uniformly site/spatial.")

        # Compute score by different methods
        ref_, com_, norm_, uncert_ = cmp.rename_dims(ref_, com_, norm_, uncert_)
        bias = com_ - ref_
        if self.method == "Collier2018":
            score = np.exp(-(np.abs(bias) - uncert_).clip(0) / norm_)
        elif self.method == "RegionalQuantiles":
            norm = quantile_map.interp(
                lat=bias["lat"], lon=bias["lon"], method="nearest"
            )
            score = (1 - (np.abs(bias) - uncert_).clip(0) / norm).clip(0, 1)
        else:
            msg = (
                "The method used to score the bias must be 'Collier2018' or "
                f"'RegionalQuantiles' but found {self.method=}"
            )
            raise ValueError(msg)
        score.attrs["units"] = 1

        # Build output datasets
        ref_out = ref_mean.to_dataset(name="mean")
        if self.use_uncertainty:
            ref_out["uncert"] = uncert
        com_out = bias.to_dataset(name="bias")
        com_out["biasscore"] = score
        try:
            lat_name = dset.get_dim_name(com_mean, "lat")
            lon_name = dset.get_dim_name(com_mean, "lon")
            com_mean = com_mean.rename({lat_name: "lat_", lon_name: "lon_"})
        except KeyError:
            pass
        com_out["mean"] = com_mean

        # Either integrate or average depending on the input var
        def _scalar(
            var, varname, region, mean=True, weight=False
        ):  # numpydoc ignore=GL08
            da = var
            if isinstance(var, xr.Dataset):
                da = var[varname]
            if dset.is_spatial(da):
                da = dset.integrate_space(
                    da,
                    varname,
                    region=region,
                    mean=mean,
                    weight=ref_ if (self.mass_weighting and weight) else None,
                )
            elif dset.is_site(da):
                da = ilr.Regions().restrict_to_region(da, region)
                da = da.mean(dim=dset.get_dim_name(da, "site"))
            else:
                raise ValueError(f"Input is neither spatial nor site: {da}")
            da = da.pint.quantify()
            return da

        # Compute scalars over all regions
        dfs = []
        for region in self.regions:
            # Period mean
            for src, var in zip(["Reference", "Comparison"], [ref_mean, com_mean]):
                var = _scalar(var, varname, region, not self.spatial_sum)
                dfs.append(
                    [
                        src,
                        str(region),
                        analysis_name,
                        "Period Mean",
                        "scalar",
                        f"{var.pint.units:~cf}",
                        float(var.pint.dequantify()),
                    ]
                )
            # Bias
            bias_scalar = _scalar(com_out, "bias", region, True)
            dfs.append(
                [
                    "Comparison",
                    str(region),
                    analysis_name,
                    "Bias",
                    "scalar",
                    f"{bias_scalar.pint.units:~cf}",
                    float(bias_scalar.pint.dequantify()),
                ]
            )
            # Bias Score
            bias_scalar_score = _scalar(com_out, "biasscore", region, True, True)
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", "divide by zero encountered in divide", RuntimeWarning
                )
                bias_scalar_score = float(bias_scalar_score.pint.dequantify())
            dfs.append(
                [
                    "Comparison",
                    str(region),
                    analysis_name,
                    "Bias Score",
                    "score",
                    "1",
                    bias_scalar_score,
                ]
            )

        # Convert to dataframe
        dfs = pd.DataFrame(
            dfs,
            columns=[
                "source",
                "region",
                "analysis",
                "name",
                "type",
                "units",
                "value",
            ],
        )
        dfs.attrs = self.__dict__.copy()
        return dfs, ref_out, com_out

    def plots(
        self,
        df: pd.DataFrame,
        ref: xr.Dataset,
        com: dict[str, xr.Dataset],
    ) -> pd.DataFrame:
        # Some initialization
        regions = [None if r == "None" else r for r in df["region"].unique()]
        com["Reference"] = ref

        # Setup plot data
        df = plt.determine_plot_limits(com).set_index("name")
        df.loc["mean", ["cmap", "title"]] = [self.cmap, "Period Mean"]
        df.loc["bias", ["cmap", "title"]] = ["seismic", "Bias"]
        df.loc["biasscore", ["cmap", "title"]] = ["plasma", "Bias Score"]

        # Build up a dataframe of matplotlib axes
        axs = [
            {
                "name": plot,
                "title": df.loc[plot, "title"],
                "region": region,
                "source": source,
                "axis": (
                    plt.plot_map(
                        ds[plot],
                        region=region,
                        vmin=df.loc[plot, "low"],
                        vmax=df.loc[plot, "high"],
                        cmap=df.loc[plot, "cmap"],
                        title=source + " " + df.loc[plot, "title"],
                    )
                    if plot in ds
                    else pd.NA
                ),
            }
            for plot in ["mean", "bias", "biasscore"]
            for source, ds in com.items()
            for region in regions
        ]
        axs = pd.DataFrame(axs).dropna(subset=["axis"])

        return axs



"""
The ILAMB net biome production scoring methodology.

See Also
--------
ILAMBAnalysis : The abstract base class from which this derives.
"""

from typing import Any

import matplotlib.pyplot as mpl
import numpy as np
import pandas as pd
import xarray as xr

import ilamb3.plot as ilplt
from ilamb3 import dataset as dset
from ilamb3.analysis.base import ILAMBAnalysis
from ilamb3.exceptions import MissingVariable, TemporalOverlapIssue


class nbp_analysis(ILAMBAnalysis):
    """
    The ILAMB net biome production scoring methodology.

    Parameters
    ----------
    evaluation_year : int, optional
        The year at which to report a difference and score. If not given, the
        last year of the reference dataset.

    Methods
    -------
    required_variables
        What variables are required.
    __call__
        The method
    """

    def __init__(self, evaluation_year: int | None = None, **kwargs: Any):
        self.evaluation_year = evaluation_year

    def required_variables(self) -> list[str]:
        """
        Return the variable names required in this analysis.

        Returns
        -------
        list
            A list of the required variables, here always [`nbp`].

        Notes
        -----
        This analysis also accepts the variable `netAtmosLandCO2Flux`. If you are
        running this routine inside an ILAMB analysis and need to use this variable,
        register it with the model as a synonym.
        """
        return ["nbp"]

    def __call__(
        self, ref: xr.Dataset, com: xr.Dataset
    ) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
        """
        Apply the ILAMB bias methodology on the given datasets.

        Parameters
        ----------
        ref : xr.Dataset
            The reference dataset.
        com : xr.Dataset
            The comparison dataset.

        Returns
        -------
        pd.DataFrame
            A dataframe with scalar and score information from the comparison.
        xr.Dataset
            A dataset containing reference grided information from the comparison.
        xr.Dataset
            A dataset containing comparison grided information from the comparison.
        """
        # Default year to evaluate if none given
        if self.evaluation_year is None:
            self.evaluation_year = int(ref["time"][-1].dt.year)

        # Check that the comparison starts in the appropriate year
        if com["time"][0].dt.year > ref["time"][0].dt.year:
            raise TemporalOverlapIssue()
        tstart = min([t for t in com["time"] if t.dt.year == ref["time"][0].dt.year])
        tend = max([t for t in com["time"] if t.dt.year == self.evaluation_year])
        com = com.sel({"time": slice(tstart, tend)})

        # Fixes to data names and checks for required variables
        if "netAtmosLandCO2Flux" in com:
            com = com.rename_vars(dict(netAtmosLandCO2Flux="nbp"))
        if "nbp" not in com:
            msg = "`nbp` or `netAtmosLandCO2Flux` needs to be in the `com` Dataset."
            raise MissingVariable(msg)
        for var in ref:
            if not var.startswith("nbp"):
                ref = ref.drop_vars(var)
            else:
                ref[var] = -ref[var]
        com["nbp"] = -com["nbp"]

        # Integrate globally
        if dset.is_spatial(com):
            com["nbp"] = dset.integrate_space(com, "nbp")
        com.load()

        # Accumulate fluxes
        def _cumsum(ds):
            for var, da in ds.items():
                da = da.pint.quantify()
                if da.pint.units is None:
                    continue
                if not (1.0 * da.pint.units).check("[mass] / [time]"):
                    continue
                da = dset.accumulate_time(ds, var)
                da = dset.convert(da, "Pg")
                ds[var] = da
            return ds

        ref = _cumsum(ref)
        com = _cumsum(com)

        # Coarsen to annual
        ref = dset.coarsen_annual(ref)
        com = dset.coarsen_annual(com)

        # Trajectory score
        ref = ref.pint.quantify()
        out_units = f"{ref['nbp'].pint.units:~cf}"
        ref = ref.pint.dequantify()
        nbp_low = ref["nbp"]
        nbp_high = ref["nbp"]
        if "bounds" in ref["nbp"].attrs and ref["nbp"].attrs["bounds"] in ref:
            nbp_low = ref[ref["nbp"].attrs["bounds"]][:, 0]
            nbp_high = ref[ref["nbp"].attrs["bounds"]][:, 1]
        uncert = xr.DataArray(
            np.sqrt(
                (ref["nbp"] - nbp_low).values ** 2, (nbp_high - ref["nbp"]).values ** 2
            ),
            coords={"year": ref["year"]},
        )
        eps = (com["nbp"] - nbp_low).clip(0) + (nbp_high - com["nbp"]).clip(0)
        traj_score = np.exp(-eps / uncert)
        traj_score = float(traj_score.mean())

        # Difference score
        ref_val = float(ref["nbp"].sel(year=self.evaluation_year))
        try:
            com_val = float(com["nbp"].sel(year=self.evaluation_year))
        except KeyError:
            com_val = np.nan
        uncert_val = float(uncert.sel(year=self.evaluation_year))
        scale = -np.log(0.5) / 1  # outside of the uncertainty window? score < 50%
        diff_score = np.exp(-scale * np.abs(com_val - ref_val) / uncert_val)

        # Scores and scalars
        df = pd.DataFrame(
            [
                {
                    "source": "Reference",
                    "region": "None",
                    "analysis": "nbp",
                    "name": f"nbp({self.evaluation_year})",
                    "type": "scalar",
                    "units": out_units,
                    "value": ref_val,
                },
                {
                    "source": "Comparison",
                    "region": "None",
                    "analysis": "nbp",
                    "name": f"nbp({self.evaluation_year})",
                    "type": "scalar",
                    "units": out_units,
                    "value": com_val,
                },
                {
                    "source": "Comparison",
                    "region": "None",
                    "analysis": "nbp",
                    "name": f"diff({self.evaluation_year})",
                    "type": "scalar",
                    "units": out_units,
                    "value": com_val - ref_val,
                },
                {
                    "source": "Comparison",
                    "region": "None",
                    "analysis": "nbp",
                    "name": "Difference Score",
                    "type": "score",
                    "units": "1",
                    "value": diff_score,
                },
                {
                    "source": "Comparison",
                    "region": "None",
                    "analysis": "nbp",
                    "name": "Trajectory Score",
                    "type": "score",
                    "units": "1",
                    "value": traj_score,
                },
            ]
        )
        return df, ref, com

    def plots(
        self,
        df: pd.DataFrame,
        ref: xr.Dataset,
        com: dict[str, xr.Dataset],
    ) -> pd.DataFrame:
        # plot over the reference limits
        vmin = ref["nbp"].min()
        vmax = ref["nbp"].max()
        if "bounds" in ref["nbp"].attrs and ref["nbp"].attrs["bounds"] in ref:
            bnd_name = ref["nbp"].attrs["bounds"]
            vmin = ref[bnd_name].min()
            vmax = ref[bnd_name].max()
        vmin = float(vmin)
        vmax = float(vmax)
        com["Reference"] = ref
        return pd.DataFrame(
            [
                {
                    "name": "accumulation",
                    "title": "nbp_accumulation",
                    "region": None,
                    "source": None,
                    "axis": plot_accumulated_nbp(com, ref, vmin=vmin, vmax=vmax),
                }
            ]
        )


def _space_labels(
    dsd: dict[str, xr.Dataset], ymin: float, maxit: int = 10
) -> dict[str, float]:
    """
    Space out the model labels for the nbp plot using a modified Laplacian
    smoothing.
    """
    sorted_dsd = {key: float(ds.isel(year=-1)["nbp"]) for key, ds in dsd.items()}
    sorted_dsd = dict(sorted(sorted_dsd.items(), key=lambda item: item[1]))
    y = np.array([v for _, v in sorted_dsd.items()])
    for j in range(maxit):
        dy = np.abs(np.diff(y))
        if dy.min() > ymin:
            break
        update = (dy[:-1] < ymin) + (dy[1:] < ymin)
        y[1:-1] = (~update) * y[1:-1] + update * 0.5 * (y[2:] + y[:-2])
    sorted_dsd = {key: v for key, v in zip(sorted_dsd.keys(), y)}
    return sorted_dsd


def plot_accumulated_nbp(
    dsd: dict[str, xr.Dataset], ref: xr.Dataset, vmin: float, vmax: float
):
    FONT_SIZE = 16
    with mpl.rc_context({"font.size": FONT_SIZE}):
        fig = mpl.figure(figsize=(12.8, 5.8))
        ax = fig.add_subplot(1, 1, 1, position=[0.1, 0.1, 0.7, 0.85])
        data_range = vmax - vmin
        fig_height = fig.get_figheight()
        pad = 0.05 * data_range
        if "bounds" in ref["nbp"].attrs and ref["nbp"].attrs["bounds"] in ref:
            da = ref[ref["nbp"].attrs["bounds"]]
            ax.fill_between(
                ref["year"],
                da.values[:, 0],
                da.values[:, 1],
                color="k",
                alpha=0.1,
                lw=0,
            )
        y_text = _space_labels(dsd, data_range / fig_height * FONT_SIZE / 50.0)
        for key, ds in dsd.items():
            ds["nbp"].plot(
                ax=ax,
                lw=2,
                color=ilplt.get_model_color(key),
            )
            ax.text(
                ds.year[-1] + 2,
                y_text[key],
                key,
                color=ilplt.get_model_color(key),
                va="center",
                size=FONT_SIZE,
            )
        ax.text(
            0.02,
            0.95,
            "Land Source",
            transform=ax.transAxes,
            size=FONT_SIZE,
            alpha=0.5,
            va="top",
        )
        ax.text(
            0.02, 0.05, "Land Sink", transform=ax.transAxes, size=FONT_SIZE, alpha=0.5
        )
        ax.set_ylabel("[Pg]")
        ax.set_ylim(vmin - pad, vmax + pad)
        ax.spines[["top", "right"]].set_visible(False)
    return ax



import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.colors import LogNorm

import ilamb3
import ilamb3.dataset as dset
from ilamb3.regions import Regions


def coerce_to_cf_compliance(da: xr.DataArray) -> xr.DataArray:
    """Try to make the units CF-compliant using the pint/cf-xarray formatter."""
    try:
        da.attrs["units"] = f"{da.pint.quantify().pint.units:~cf}"
    except Exception:
        pass
    return da


def get_extents(da: xr.DataArray) -> list[float]:
    """Find the extent of the non-null data."""
    lat = xr.where(da.notnull(), da[dset.get_coord_name(da, "lat")], np.nan)
    lon = xr.where(da.notnull(), da[dset.get_coord_name(da, "lon")], np.nan)
    extents = [float(lon.min()), float(lon.max()), float(lat.min()), float(lat.max())]
    # if a da is all nan, then the extents cause trouble downstream in plotting.
    if np.isnan(extents).any():
        return [-180.0, 180.0, -90.0, 90.0]
    return extents


def compute_extent_area(extents):
    return (extents[1] - extents[0]) * (extents[3] - extents[2])


def compute_overlap_fracs(
    extents_a: list[float], extents_b: list[float]
) -> tuple[float, float]:
    """Return the fractions of overlap."""

    overlap = [op(a, b) for op, a, b in zip([max, min, max, min], extents_a, extents_b)]
    area_O = (
        compute_extent_area(overlap)
        if (
            (extents_a[0] < extents_b[1])
            & (extents_a[1] > extents_b[0])
            & (extents_a[2] < extents_b[3])
            & (extents_a[3] > extents_b[2])
        )
        else 0.0
    )
    area_A = compute_extent_area(extents_a)
    area_B = compute_extent_area(extents_b)
    frac_A = area_O / area_A if area_A > 0 else 0.0
    frac_B = area_O / area_B if area_B > 0 else 0.0
    return frac_A, frac_B


def pick_projection(
    extents: list[float], fraction_threshold: float = 0.85
) -> tuple[ccrs.Projection, float]:
    """Given plot extents choose projection and aspect ratio."""
    lon = ilamb3.conf["plot_central_longitude"]
    if compute_overlap_fracs([-180, 180, 60, 90], extents)[1] > fraction_threshold:
        return ccrs.Orthographic(central_latitude=+90, central_longitude=lon), 1.0
    if compute_overlap_fracs([-180, 180, -90, -60], extents)[1] > fraction_threshold:
        return ccrs.Orthographic(central_latitude=-90, central_longitude=lon), 1.0
    if compute_overlap_fracs([-125, -66.5, 20, 50], extents)[1] > fraction_threshold:
        return ccrs.LambertConformal(), 2.05  # USA
    if compute_extent_area(extents) / compute_extent_area([-180, 180, -90, 90]) > 0.5:
        return ccrs.Robinson(central_longitude=lon), 2.0  # Global
    # If none of above, use cyclindrical
    aspect_ratio = max(extents[1], extents[0]) - min(extents[1], extents[0])
    aspect_ratio /= max(extents[3], extents[2]) - min(extents[3], extents[2])
    proj = ccrs.PlateCarree(central_longitude=np.array(extents)[:2].mean())
    return proj, aspect_ratio


def finalize_plot(ax: plt.Axes, extents: list[float]) -> plt.Axes:
    """Add some final features to our plots."""
    ax.add_feature(
        cfeature.NaturalEarthFeature(
            "physical", "land", "110m", edgecolor="face", facecolor="0.875"
        ),
        zorder=-1,
    )
    ax.add_feature(
        cfeature.NaturalEarthFeature(
            "physical", "ocean", "110m", edgecolor="face", facecolor="0.750"
        ),
        zorder=-1,
    )
    # cleanup plotting extents
    percent_pad = 0.1
    if (extents[1] - extents[0]) > 300:
        extents[:2] = [-180, 180]  # set_extent doesn't like (0,360)
        extents[2:] = [-90, 90]
    else:
        dx = percent_pad * (extents[1] - extents[0])
        dy = percent_pad * (extents[3] - extents[2])
        extents = [
            max(extents[0] - dx, -180),
            min(extents[1] + dx, 180),
            max(extents[2] - dy, -90),
            min(extents[3] + dy, 90),
        ]
    ax.set_extent(extents, ccrs.PlateCarree())
    return ax


def plot_map(da: xr.DataArray, **kwargs):
    # Process some options
    ncolors = kwargs.pop("ncolors") if "ncolors" in kwargs else 9
    ticklabels = kwargs.pop("ticklabels") if "ticklabels" in kwargs else None
    kwargs["cmap"] = plt.get_cmap(
        kwargs["cmap"] if "cmap" in kwargs else "viridis", ncolors
    )
    title = kwargs.pop("title") if "title" in kwargs else ""
    da = coerce_to_cf_compliance(da)

    # Process region if given
    ilamb_regions = Regions()
    region = kwargs.pop("region") if "region" in kwargs else None
    da = ilamb_regions.restrict_to_region(da, region)

    # Setup figure and its projection
    extents = get_extents(da)
    proj, aspect = pick_projection(extents)
    figsize = kwargs.pop("figsize") if "figsize" in kwargs else (6 * 1.03, 6 / aspect)
    _, ax = plt.subplots(
        dpi=ilamb3.conf["figure_dpi"],
        tight_layout=(kwargs.pop("tight_layout") if "tight_layout" in kwargs else True),
        figsize=figsize,
        subplot_kw={"projection": proj},
    )

    # Setup colorbar arguments
    cba = {"label": da.attrs["units"]}
    if "cbar_kwargs" in kwargs:
        cba.update(kwargs.pop("cbar_kwargs"))

    # We can't make temporal maps
    if dset.is_temporal(da):
        raise ValueError("Cannot make spatio-temporal plots")

    # Space or sites?
    if dset.is_spatial(da):
        out_plot = da.plot(
            ax=ax, transform=ccrs.PlateCarree(), cbar_kwargs=cba, **kwargs
        )
        if ticklabels is not None:
            out_plot.colorbar.set_ticklabels(ticklabels)
    elif dset.is_site(da):
        out_plot = xr.plot.scatter(
            da.to_dataset(),
            x=dset.get_coord_name(da, "lon"),
            y=dset.get_coord_name(da, "lat"),
            hue=da.name,
            lw=0,
            s=15,
            cbar_kwargs=cba,
            transform=ccrs.PlateCarree(),
            **kwargs,
        )
        if ticklabels is not None:
            out_plot.colorbar.set_ticklabels(ticklabels)
    else:
        raise ValueError("plotting error")
    ax.set_title(title)
    ax = finalize_plot(ax, extents)
    return ax


def _pick_convert_calendar_align_on_option(da: xr.DataArray) -> None | str:
    """
    Resturn the best option based on guidance in the xarray documentation.
    """
    t = da[dset.get_dim_name(da, "time")]
    if (hasattr(t, "dt") and t.dt.calendar == "360_day") or (
        "calendar" in t.attrs and t.attrs["calendar"] == "360_day"
    ):
        dt = dset.get_mean_time_frequency(da)
        if dt < 1:
            return "year"
        else:
            return "date"
    return None


def plot_curve(dsd: dict[str, xr.Dataset], varname: str, **kwargs):
    # Parse some options
    vmin = kwargs.pop("vmin") if "vmin" in kwargs else None
    vmax = kwargs.pop("vmax") if "vmax" in kwargs else None
    xticks = kwargs.pop("xticks") if "xticks" in kwargs else None
    xticklabels = kwargs.pop("xticklabels") if "xticklabels" in kwargs else None
    title = kwargs.pop("title") if "title" in kwargs else ""
    ylabel = kwargs.pop("ylabel") if "ylabel" in kwargs else None

    # Setup figure
    ASPECT = 1.618
    figsize = kwargs.pop("figsize") if "figsize" in kwargs else (6, 6 / ASPECT)
    _, ax = plt.subplots(
        dpi=ilamb3.conf["figure_dpi"],
        tight_layout=(kwargs.pop("tight_layout") if "tight_layout" in kwargs else True),
        figsize=figsize,
    )

    # Convert to single calendar for plotting
    dad = {
        source: (
            ds[varname].convert_calendar(
                "noleap", align_on=_pick_convert_calendar_align_on_option(ds[varname])
            )
            if "time" in ds[varname].dims
            else ds[varname]
        )
        for source, ds in dsd.items()
    }

    # Coerce units
    dad = {source: coerce_to_cf_compliance(da) for source, da in dad.items()}

    # Plot curves
    ref = dad.pop("Reference")
    ref.plot(ax=ax, color="k", label="Reference")
    for source, da in dad.items():
        da.plot(ax=ax, color=get_model_color(source), label=source)

    ax.legend()
    ax.set_title(title)
    if xticks is not None:
        ax.set_xticks(xticks)
    if xticklabels is not None:
        ax.set_xticklabels(xticklabels)
    if vmin is not None and vmax is not None:
        ax.set_ylim(vmin, vmax)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    return ax


def plot_distribution(da: xr.DataArray, **kwargs):
    da = coerce_to_cf_compliance(da)
    _, ax = plt.subplots(
        dpi=ilamb3.conf["figure_dpi"], tight_layout=True, figsize=(6, 5.25)
    )
    da.plot(
        ax=ax,
        norm=LogNorm(),
        vmin=1e-4,
        vmax=1e-1,
        cmap="plasma",
        cbar_kwargs={"label": "Fraction of total data"},
    )
    ax.set_title(kwargs.pop("title") if "title" in kwargs else "")
    return ax


def plot_response(
    ref_mean: xr.DataArray,
    ref_std: xr.DataArray,
    mean: xr.DataArray,
    std: xr.DataArray,
    comparison_name: str,
    **kwargs,
):
    def _plot(ax, m, s, c="k", lbl="Reference"):
        ax.fill_between(
            m[m.dims[0]].values,
            m - s,
            m + s,
            color=c,
            alpha=0.1,
            lw=0,
            label=f"{lbl} variability",
        )
        m.plot(ax=ax, color=c, label=f"{lbl} mean")
        return ax

    _, ax = plt.subplots(
        dpi=ilamb3.conf["figure_dpi"], tight_layout=True, figsize=(6, 5.25)
    )
    ax = _plot(ax, ref_mean, ref_std)
    ax = _plot(ax, mean, std, "r", comparison_name)
    ax.legend()
    ax.set_title(kwargs.pop("title") if "title" in kwargs else "")
    return ax


def plot_taylor_diagram(df: pd.DataFrame):
    """
    Create a Taylor diagram.

    This is adapted from the code by Yannick Copin found here:

    https://gist.github.com/ycopin/3342888
    """
    import mpl_toolkits.axisartist.floating_axes as FA
    import mpl_toolkits.axisartist.grid_finder as GF
    from matplotlib.projections import PolarAxes

    # correlation ticks and labels
    rlocs = np.array([0.1, 0.3, 0.5, 0.7, 0.9])
    tlocs = np.arccos(rlocs)
    gl1 = GF.FixedLocator(tlocs)
    tf1 = GF.DictFormatter(dict(zip(tlocs, map(str, rlocs))))

    # add a curvilinear grid helper
    smax = max(
        2, 1.1 * df[df["name"] == "Normalized Standard Deviation [1]"]["value"].max()
    )
    tr = PolarAxes.PolarTransform()
    ghelper = FA.GridHelperCurveLinear(
        tr,
        extremes=(0, np.pi / 2, 0, smax),
        grid_locator1=gl1,
        tick_formatter1=tf1,
    )

    # create figure
    with mpl.rc_context({"font.size": 13}):
        fig = plt.figure(dpi=ilamb3.conf["figure_dpi"], figsize=(6, 6))
        ax = FA.FloatingSubplot(fig, 111, grid_helper=ghelper)
        fig.add_subplot(ax)
        ax.axis["top"].set_axis_direction("bottom")
        ax.axis["top"].toggle(ticklabels=True, label=True)
        ax.axis["top"].major_ticklabels.set_axis_direction("top")
        ax.axis["top"].label.set_axis_direction("top")
        ax.axis["top"].label.set_text("Correlation")
        ax.axis["left"].set_axis_direction("bottom")
        ax.axis["left"].label.set_text("Normalized standard deviation")
        ax.axis["right"].set_axis_direction("top")
        ax.axis["right"].toggle(ticklabels=True)
        ax.axis["right"].major_ticklabels.set_axis_direction("left")
        ax.axis["bottom"].set_visible(False)
        ax.grid(True)

        # plot data
        ax = ax.get_aux_axes(tr)
        for source, grp in df[
            df["name"].isin(["Normalized Standard Deviation [1]", "Correlation [1]"])
        ].groupby("source"):
            std = grp[grp["name"] == "Normalized Standard Deviation [1]"].iloc[0][
                "value"
            ]
            corr = grp[grp["name"] == "Correlation [1]"].iloc[0]["value"]
            ax.plot(
                np.arccos(corr.clip(-1, 1)),
                std,
                "o",
                color=get_model_color(source),
                mew=0,
                ms=8,
            )

        # Add reference point and stddev contour
        ax.plot([0], 1, "k*", ms=12, mew=0)
        t = np.linspace(0, np.pi / 2)
        r = np.zeros_like(t) + 1
        ax.plot(t, r, "k--")

        # centralized rms contours
        rs, ts = np.meshgrid(np.linspace(0, smax), np.linspace(0, np.pi / 2))
        rms = np.sqrt(1 + rs**2 - 2 * rs * np.cos(ts))
        contours = ax.contour(ts, rs, rms, 5, colors="k", alpha=0.4)
        ax.clabel(contours, fmt="%1.1f")

    return ax


def determine_plot_limits(
    dsd: xr.Dataset | dict[str, xr.Dataset],
    percent_pad: float = 1.0,
    symmetrize: list[str] = ["bias"],
) -> pd.DataFrame:
    """Return a dataframe with the plot minima and maxima."""
    if isinstance(dsd, xr.Dataset):
        dsd = {"key": xr.Dataset}
    plots = set([var for _, ds in dsd.items() for var in ds.data_vars])
    out = []
    for plot in plots:
        # Score maps are always [0,1]
        if "score" in plot:
            out.append({"name": plot, "low": 0, "high": 1})
            continue
        # Which datasets have this plot, the data type must be a float
        plot_ds = {
            key: ds
            for key, ds in dsd.items()
            if plot in ds and (np.issubdtype(ds[plot].dtype, np.floating))
        }
        if not plot_ds:
            continue
        # Stack the data together into a single array
        data = np.hstack(
            [
                np.ma.masked_invalid(ds[plot].to_numpy()).compressed()
                for _, ds in plot_ds.items()
            ]
        )
        # If we have data, then take the quantiles, otherwise it doesn't matter.
        if data.size > 0:
            data = np.quantile(data, [percent_pad * 0.01, 1 - percent_pad * 0.01])
        else:
            data = np.asarray([0.0, 1.0])
        # Symmetrize
        if [tag for tag in symmetrize if tag in plot]:
            vmax = max(np.abs(data))
            data[0] = -vmax
            data[1] = vmax
        out.append({"name": plot, "low": data[0], "high": data[1]})
    return pd.DataFrame(out)


def get_model_color(
    model: str, base_cmap: str = "rainbow"
) -> tuple[float, float, float, float]:
    if model == "Reference":
        return (0.0, 0.0, 0.0, 1.0)
    MODEL_PREFIXES = np.array(
        [
            "ACC",
            "AWI",
            "BCC",
            "CAM",
            "CAS",
            "CES",
            "CIE",
            "CMC",
            "CNR",
            "CAN",
            "E3S",
            "EC",
            "FGO",
            "FIO",
            "GFD",
            "GIS",
            "HAD",
            "ICO",
            "IIT",
            "INM",
            "IPS",
            "KAC",
            "KIO",
            "MCM",
            "MIR",
            "MPI",
            "MRI",
            "NES",
            "Nor",
            "SAM",
            "TAI",
            "UKE",
        ]
    )
    cmap = plt.get_cmap(base_cmap, len(MODEL_PREFIXES))
    return cmap(MODEL_PREFIXES.searchsorted(model.upper()))



import pytest

from ilamb3.exceptions import AnalysisFailure


def test_exceptions():
    with pytest.raises(AnalysisFailure, match=r".*my_analysis.*my_model.*"):
        raise AnalysisFailure("my_analysis", "my_model")



"""Configuration for ilamb3"""

import contextlib
import copy
import os
from pathlib import Path

import yaml

import ilamb3.regions as reg

defaults = {
    "regions": [None],
    "prefer_regional_quantiles": False,
    "quantile_database": "quantiles/quantiles_Whittaker_cmip5v6.parquet",
    "quantile_threshold": 70,
    "use_uncertainty": False,
    "model_name_facets": ["source_id", "member_id", "grid_label"],
    "plot_central_longitude": 0,
    "comparison_groupby": ["source_id", "member_id", "grid_label"],
    "use_cached_results": False,
    "figure_dpi": 100,
    "debug_mode": False,
    "run_mode": "interactive",  # for internal use
}


class Config(dict):
    """A global configuration object used in the package."""

    def __init__(self, filename: Path | None = None, **kwargs):
        if filename is not None:
            self.filename = Path(filename)
        elif "ILAMB3_CONFIGURATION" in os.environ:
            self.filename = Path(os.environ["ILAMB3_CONFIGURATION"])
        else:
            self.filename = Path.home() / ".config/ilamb3/conf.yaml"
        self.filename.parent.mkdir(parents=True, exist_ok=True)
        self.reload_all()
        self.temp = None
        super().__init__(**kwargs)

    def __repr__(self):
        return yaml.dump(dict(self))

    def reset(self):
        """Return to defaults."""
        self.clear()
        self.update(copy.deepcopy(defaults))

    def save(self, filename: Path | None = None):
        """Save current configuration to file as YAML."""
        filename = filename or self.filename
        filename.parent.mkdir(parents=True, exist_ok=True)
        with open(filename, "w") as f:
            yaml.dump(dict(self), f)

    @contextlib.contextmanager
    def _unset(self, temp):
        yield
        self.clear()
        self.update(temp)

    def set(
        self,
        *,
        build_dir: str | None = None,
        regions: list[str] | None = None,
        prefer_regional_quantiles: bool | None = None,
        use_uncertainty: bool | None = None,
        model_name_facets: list[str] | None = None,
        plot_central_longitude: float | None = None,
        comparison_groupby: list[str] | None = None,
        use_cached_results: bool | None = None,
        figure_dpi: int | None = None,
        debug_mode: bool | None = None,
    ):
        """Change ilamb3 configuration options."""
        temp = copy.deepcopy(self)
        if build_dir is not None:
            self["build_dir"] = str(build_dir)
        if regions is not None:
            ilamb_regions = reg.Regions()
            does_not_exist = set(regions) - set(ilamb_regions._regions) - set([None])
            if does_not_exist:
                raise ValueError(
                    f"Cannot run ILAMB over these regions {list(does_not_exist)} which are not registered in our system {list(ilamb_regions._regions)}"
                )
            self["regions"] = regions
        if prefer_regional_quantiles is not None:
            self["prefer_regional_quantiles"] = bool(prefer_regional_quantiles)
        if use_uncertainty is not None:
            self["use_uncertainty"] = bool(use_uncertainty)
        if model_name_facets is not None:
            self["model_name_facets"] = model_name_facets
        if plot_central_longitude is not None:
            self["plot_central_longitude"] = float(plot_central_longitude)
        if comparison_groupby is not None:
            self["comparison_groupby"] = comparison_groupby
        if use_cached_results is not None:
            self["use_cached_results"] = bool(use_cached_results)
        if figure_dpi is not None:
            self["figure_dpi"] = int(figure_dpi)
        if debug_mode is not None:
            self["debug_mode"] = bool(debug_mode)
        return self._unset(temp)

    def __getitem__(self, item):
        if item in self:
            return super().__getitem__(item)
        elif item in defaults:
            return defaults[item]
        else:
            raise KeyError(item)

    def get(self, key, default=None):
        if key in self:
            return super().__getitem__(key)
        return default

    def reload_all(self):
        self.reset()
        self.load()

    def load(self, filename: Path | None = None):
        """Update global config from YAML file or default file if None."""
        filename = filename or self.filename
        if filename.is_file():
            with open(filename) as f:
                try:
                    self.update(yaml.safe_load(f))
                except Exception:
                    pass


conf = Config()
conf.reload_all()



from pathlib import Path

import numpy as np
import xarray as xr

import ilamb3
from ilamb3.regions import Regions
from ilamb3.tests.test_dataset import generate_test_dset


def test_basic():
    reg = Regions()
    assert reg.get_name("euro") == "Europe"
    assert reg.get_source("euro") == "Global Fire Emissions Database (GFED)"


def test_netcdf():
    # can we add regions via a dataset?
    cat = ilamb3.ilamb_catalog()
    dsr = xr.open_dataset(cat.fetch("regions/GlobalLand.nc"))
    reg = Regions()
    lbl = reg.add_netcdf(dsr)
    assert "global" in lbl

    # can we add regions via a netcdf file?
    dsr.to_netcdf("tmp.nc")
    lbl = reg.add_netcdf("tmp.nc")
    assert "global" in lbl

    # does the region work?
    ds = generate_test_dset()
    ds = reg.restrict_to_region(ds, "global")
    assert np.isclose(ds["da"].mean(), 4.28285108e-09)

    Path("tmp.nc").unlink()



"""
ILAMB transforms for selecting data.
"""

from typing import Any

import xarray as xr

import ilamb3.dataset as dset
from ilamb3.transform.base import ILAMBTransform


class select_depth(ILAMBTransform):
    """
    Select a depth if the dimension exists.

    Parameters
    ----------
    value : float
        The value at which to select the nearest depth.
    vmin : float
        The minimum depth to slice.
    vmax : float
        The maximum depth to slice.

    Note
    ----
    Either `value` should be specified or `vmin` and `vmax` but not both.
    """

    def __init__(
        self,
        value: float | None = None,
        vmin: float | None = None,
        vmax: float | None = None,
        **kwargs: Any,
    ):
        self.value = value
        self.vmin = vmin
        self.vmax = vmax
        ok = True
        if value is None:
            ok = vmin is not None and vmax is not None
        if vmin is None:
            ok = vmax is None and value is not None
        if vmax is None:
            ok = vmin is None and value is not None
        if not ok:
            raise ValueError(
                "This transform allows for a single depth selection with `value` or a range with "
                f"both `vmin` and `vmax`, but I found {value=} {vmin=} {vmax=} and these other keywords {kwargs}"
            )

    def required_variables(self) -> list[str]:
        """
        Return the variables this transform uses.
        """
        return []

    def __call__(self, ds: xr.Dataset) -> xr.Dataset:
        """
        Select a depth or depth range of the input dataset, if a depth dimension is found.
        """
        if not dset.is_layered(ds):
            return ds
        depth_name = dset.get_dim_name(ds, "depth")
        if self.value is not None:
            return ds.sel({depth_name: self.value}, method="nearest", drop=True)
        else:
            return ds.sel({depth_name: slice(self.vmin, self.vmax)})
        return ds



"""A library of functions used to benchmark earth system models."""

# import these packages so that units via pint will be possible once anything
# from ilamb is imported.
import importlib
import warnings

import pint_xarray  # noqa
import pooch
import xarray as xr
from cf_xarray.units import units

from ilamb3._version import __version__  # noqa
from ilamb3.config import conf  # noqa

# additional units that pint/cf-xarray does not handle
units.define("kg = 1e3 * g")
units.define("Mg = 1e6 * g")
units.define("Pg = 1e15 * g")
units.define("Sv = 1e6 m**3 / s")

# we don't really have data versions for the collection :/
ILAMB_DATA_VERSION = "0.1"

# remove nuissance warnings from multiple fill values
warnings.simplefilter("ignore", category=xr.SerializationWarning)


def ilamb_catalog() -> pooch.Pooch:
    """
    Return the pooch ilamb reference data catalog.

    Returns
    -------
    pooch.Pooch
        The intake ilamb reference data catalog.
    """
    registry = pooch.create(
        path=pooch.os_cache("ilamb3"),
        base_url="https://www.ilamb.org/ILAMB-Data/DATA",
        version=ILAMB_DATA_VERSION,
        env="ILAMB_ROOT",
    )
    registry.load_registry(
        importlib.resources.open_binary("ilamb3.registry", "ilamb.txt")
    )
    return registry


def iomb_catalog() -> pooch.Pooch:
    """
    Return the pooch iomb reference data catalog.

    Returns
    -------
    pooch.Pooch
        The intake ilamb reference data catalog.
    """
    registry = pooch.create(
        path=pooch.os_cache("ilamb3"),
        base_url="https://www.ilamb.org/ilamb3-data",
        version=ILAMB_DATA_VERSION,
        env="ILAMB_ROOT",
    )
    registry.load_registry(
        importlib.resources.open_binary("ilamb3.registry", "iomb.txt")
    )
    return registry


def test_catalog() -> pooch.Pooch:
    """
    Return the pooch iomb reference data catalog.

    Returns
    -------
    pooch.Pooch
        The intake ilamb reference data catalog.
    """
    registry = pooch.create(
        path=pooch.os_cache("ilamb3"),
        base_url="https://www.ilamb.org/IOMB-Data/DATA",
        version=ILAMB_DATA_VERSION,
        env="ILAMB_ROOT",
    )
    registry.load_registry(
        importlib.resources.open_binary("ilamb3.registry", "test.txt")
    )
    return registry


__all__ = [
    "dataset",
    "compare",
    "analysis",
    "regions",
    "run",
    "ilamb_catalog,",
    "conf",
    "ILAMB_DATA_VERSION",
]
xr.set_options(keep_attrs=True)



"""Region definitions for use in the ILAMB system."""

import os

import numpy as np
import xarray as xr

import ilamb3.dataset as dset


def restrict_to_bbox(
    da: xr.DataArray, lat0: float, latf: float, lon0: float, lonf: float
):
    """Return the dataarray selected to the nearest bounding box.

    This is awkward because as of `v2023.6.0`, the `method` keyword cannot be used in
    slices. Note that this routine will sort the dimensions because slicing does not
    work well on unsorted indices.
    """
    assert isinstance(da, xr.DataArray)
    lat_name = dset.get_coord_name(da, "lat")
    lon_name = dset.get_coord_name(da, "lon")
    if dset.is_site(da):
        site_name = dset.get_dim_name(da, "site")
        da = da.sel(
            {
                site_name: (
                    (da[lat_name] >= lat0)
                    & (da[lat_name] <= latf)
                    & (da[lon_name] >= lon0)
                    & (da[lon_name] <= lonf)
                )
            }
        )
    else:
        da = da.sel(
            {
                lat_name: slice(
                    da[lat_name].sel({lat_name: lat0}, method="nearest"),
                    da[lat_name].sel({lat_name: latf}, method="nearest"),
                ),
                lon_name: slice(
                    da[lon_name].sel({lon_name: lon0}, method="nearest"),
                    da[lon_name].sel({lon_name: lonf}, method="nearest"),
                ),
            }
        )
    return da


def restrict_to_region(da: xr.DataArray, dar: xr.DataArray):
    """."""
    assert isinstance(da, xr.DataArray)
    lat_name = dset.get_coord_name(da, "lat")
    lon_name = dset.get_coord_name(da, "lon")
    rlat_name = dset.get_dim_name(dar, "lat")
    rlon_name = dset.get_dim_name(dar, "lon")
    dar = dar.rename({rlat_name: lat_name, rlon_name: lon_name})
    return restrict_to_bbox(
        xr.where(
            dar.interp(
                {lat_name: da[lat_name], lon_name: da[lon_name]},
                method="nearest",
            ),
            da,
            np.nan,
        ),
        dar[lat_name].min(),
        dar[lat_name].max(),
        dar[lon_name].min(),
        dar[lon_name].max(),
    )


class Regions:
    """A class for unifying the treatment of regions in ILAMB."""

    _regions = {}
    _sources = {}

    @property
    def regions(self):
        """Return a list of region identifiers."""
        return Regions._regions.keys()

    def add_latlon_bounds(
        self,
        label: str,
        name: str,
        lats: list[float],
        lons: list[float],
        source: str = "user-provided latlon bounds",
    ) -> None:
        """Add a region by lat/lon bounds.

        Parameters
        ----------
        label
            The label by which the region will be known in the ilamb system. Note that
            this will overwrite any current region known by this label.
        name
            The region text displayed in the ilamb output.
        lats
            The lower and upper latitude bounds [-90,90] defining the region.
        lons
            The lower and upper longitude bounds [-180,180] defining the region.
        source
            An optional text description for the source of these regions.

        """
        assert len(lats) == 2
        assert len(lons) == 2
        rtype = 0
        Regions._regions[label] = [rtype, name, lats, lons]
        Regions._sources[label] = source

    def add_netcdf(self, netcdf: str | xr.Dataset) -> list[str]:
        """Add regions found in a netCDF file or dataset.

        Region formatting guidelines can be found
        [here](https://www.ilamb.org/doc/custom_regions.html).

        Parameters
        ----------
        netcdf
            The filename or dataset which contain the regions.

        Returns
        -------
        labels
            The list of region labels added to the ilamb system as a result of this
            function call.

        """
        rtype = 1
        if isinstance(netcdf, str):
            dsr = xr.open_dataset(netcdf)
        else:
            dsr = netcdf
        if "ids" in dsr:
            ids = "ids"
        else:
            ids = [
                v
                for v in dsr.data_vars
                if dsr[v].ndim == 2 and dsr[v].dtype == np.integer
            ]
            if len(ids) == 0:
                raise ValueError(
                    f"Found no 2d integer arrays in the region file {netcdf}"
                )
            if len(ids) > 1:
                raise ValueError(
                    f"Amiguous integer array for regions in {netcdf}: {ids}"
                )
            ids = ids[0]
        labels = list(dsr[dsr[ids].attrs["labels"]].to_numpy())
        names = list(dsr[dsr[ids].attrs["names"]].to_numpy())
        for label, name in zip(labels, names):
            dar = xr.where(dsr[ids] == labels.index(label), 1, 0)
            Regions._regions[label] = [rtype, name, dar]
            Regions._sources[label] = (
                os.path.basename(netcdf) if isinstance(netcdf, str) else "dataset"
            )
        return labels

    def get_name(self, label: str) -> str:
        """Return the region name given its label."""
        return Regions._regions[label][1]

    def get_source(self, label: str):
        """Return the source of the region given its label."""
        return Regions._sources[label]

    def restrict_to_region(
        self,
        var: xr.Dataset | xr.DataArray,
        label: str | None,
    ) -> xr.Dataset | xr.DataArray:
        """Given the region label and a variable, return a mask.

        ILAMB intermediate netCDF files can contain 2 grids defined by (lat,lon) and
        (lat_,lon_, the composed grid). For this reason, we need to restrict to regions
        one dataarray at a time.
        """
        if label is None:
            return var
        rdata = Regions._regions[label]
        rtype = rdata[0]
        if rtype == 0:
            _, _, rlat, rlon = rdata
            if isinstance(var, xr.DataArray):
                out = restrict_to_bbox(var, rlat[0], rlat[1], rlon[0], rlon[1])
            else:
                out = {}
                for key in var:
                    try:
                        out[key] = restrict_to_bbox(
                            var[key], rlat[0], rlat[1], rlon[0], rlon[1]
                        )
                    except KeyError:
                        continue
                out = xr.Dataset(out)
            return out
        if rtype == 1:
            _, _, dar = rdata
            if isinstance(var, xr.DataArray):
                out = restrict_to_region(var, dar)
                out.name = var.name
            else:
                out = {}
                for key in var:
                    try:
                        out[key] = restrict_to_region(var[key], dar)
                    except KeyError:
                        continue
                out = xr.Dataset(out)
            return out
        raise ValueError(f"Region type #{rtype} not recognized")

    def region_scalars_to_map(self, scalars: dict[str, float]) -> xr.DataArray:
        # check that regions are part of our system
        diff = set(scalars) - set(Regions._regions)
        if diff:
            raise ValueError(
                f"Keys in the scalar dictionary aren't registered regions: {diff}"
            )
        # make sure all regions come from the same source (and therefore grid)
        sources = [Regions._sources[r] for r in scalars]
        assert all([src == sources[0] for src in sources])
        # build up region map depending on region type
        rtype = [Regions._regions[r][0] for r in scalars][0]
        if rtype == 0:
            da = xr.DataArray(
                data=np.ones((180, 360)) * np.nan,
                dims=["lat", "lon"],
                coords=dict(
                    lat=np.linspace(-89.5, 89.5, 180),
                    lon=np.linspace(-179.5, 179.5, 360),
                ),
            )
            for r, val in scalars.items():
                da = xr.where(
                    (da["lat"] > Regions._regions[r][2][0])
                    & (da["lat"] <= Regions._regions[r][2][1])
                    & (da["lon"] > Regions._regions[r][3][0])
                    & (da["lon"] <= Regions._regions[r][3][1]),
                    val,
                    da,
                )
        elif rtype == 1:
            da = [
                xr.where(Regions._regions[r][2] == 0, np.nan, 1) * val
                for r, val in scalars.items()
            ]
            da = xr.concat(da, dim="region")
            mask = da.isnull().all(dim="region")
            da = da.sum(dim="region")
            da = xr.where(mask, np.nan, da)
        else:
            raise NotImplementedError("Region type not implemented.")
        return da


# If no regions have been registered, then add these
if len(Regions().regions) == 0:
    r = Regions()
    regions = """
    bona, Boreal North America             , 50, 80,-170, -60
    tena, Temperate North America          , 30, 50,-125, -66
    ceam, Central America                  , 10, 30,-115, -80
    nhsa, Northern Hemisphere South America,  0, 13, -80, -50
    shsa, Southern Hemisphere South America,-60,  0, -80, -33
    euro, Europe                           , 35, 70, -10,  30
    mide, Middle East                      , 20, 40, -10,  60
    nhaf, Northern Hemisphere Africa       ,  0, 20, -20,  45
    shaf, Southern Hemisphere Africa       ,-35,  0,  10,  45
    boas, Boreal Asia                      , 55, 70,  30, 180
    ceas, Central Asia                     , 30, 55,  30, 143
    seas, Southeast Asia                   ,  5, 30,  65, 120
    eqas, Equatorial Asia                  ,-10, 10, 100, 150
    aust, Australia                        ,-41,-11, 112, 154
    """.strip().split(
        "\n"
    )
    for line in regions:
        lbl, name, lat0, latf, lon0, lonf = line.split(",")
        r.add_latlon_bounds(
            lbl.strip(),
            name.strip(),
            [float(lat0), float(latf)],
            [float(lon0), float(lonf)],
            "Global Fire Emissions Database (GFED)",
        )



from pathlib import Path

import ilamb3


def test_ilamb():
    reg = ilamb3.ilamb_catalog()
    fname = reg.fetch("test/Site/tas.nc")
    assert str(reg.abspath).endswith(ilamb3.ILAMB_DATA_VERSION)
    assert Path(fname).is_file()


def test_iomb():
    reg = ilamb3.iomb_catalog()
    fname = reg.fetch(
        "RAPID-2023-1a/obs4MIPs_NOC_RAPID-2023-1a_mon_msftmz_gm_v20250902.nc"
    )
    assert str(reg.abspath).endswith(ilamb3.ILAMB_DATA_VERSION)
    assert Path(fname).is_file()


def test_test():
    reg = ilamb3.test_catalog()
    fname = reg.fetch("test/thetao.nc")
    assert str(reg.abspath).endswith(ilamb3.ILAMB_DATA_VERSION)
    assert Path(fname).is_file()



"""
An ILAMB transform for computing the depth gradient.
"""

import xarray as xr

import ilamb3.dataset as dset
from ilamb3.transform.base import ILAMBTransform


class depth_gradient(ILAMBTransform):
    """
    Compute the depth gradient of the dataset using least squares regression.
    """

    def __init__(self):
        pass

    def required_variables(self) -> list[str]:
        """Return the variables this transform requires, none in this case."""
        return []

    def __call__(self, ds: xr.Dataset) -> xr.Dataset:
        """
        Return the depth gradient of the input dataset, if a depth dimension
        exists.

        Note
        ----
        This transform will also integrate out the temporal dimension. It may be
        useful to use this transform in conjunction with the `select_depth` to
        pick a range of depths across which you want to estimate the gradient.
        """
        if not dset.is_layered(ds):
            return ds
        for var, da in ds.items():
            if dset.is_temporal(da):
                ds[var] = dset.integrate_time(ds, var, mean=True)
        grad = ds.polyfit(dset.get_dim_name(ds, "depth"), deg=1).sel(
            degree=1, drop=True
        )
        grad = grad.rename_vars(
            {
                var: var.replace("_polyfit_coefficients", "_depth_gradient")
                for var in grad
            }
        )
        for var, da in grad.items():
            if "_depth_gradient" in var:
                orig = var.replace("_depth_gradient", "")
                grad[var].attrs["units"] = f"{ds[orig].attrs['units']} m-1"
        return grad
